//-----------------------------------------------------------------------------
//
//  $Logfile:: /game/specialfx.h                                              $
// $Revision:: 2                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 6:41 PM                                               $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// special effects
//

#ifndef __SPECIAL_FX_H__
#define __SPECIAL_FX_H__

#include "g_local.h"
#include "entity.h"
#include "trigger.h"
#include "light.h"
#include "scriptslave.h"

void SpawnBlastDamage( trace_t* trace, int damage, Entity* attacker );
void Particles( Vector org, Vector norm, int count, int lightstyle, int flags );
void SpawnBlood( Vector org, Vector vel, int damage );
void SpawnSparks( Vector org,   Vector norm, int count );
void BurnWall( Vector org, Vector end, int amount );
void SpawnRocketExplosion( Vector org );
void SpawnScaledExplosion( Vector org, float scale );
void SpawnTeleportEffect( Vector org, int lightstyle );
void SpawnBeam( Vector start, Vector end, int parent_entnum, int modelindex, float alpha, float life, int flags );

void SpawnTempDlight( Vector org, float r, float g, float b, float radius, float decay, float life );

void TempModel( Entity* parent, Vector origin, Vector angles, const char* modelname, int anim, float scale, float alpha, int flags, float life );

void TesselateModel( Entity* ent, int min_size, int max_size, Vector dir, float power, float percentage, int thickness, Vector origin, int type = TESS_DEFAULT_TYPE, int lightstyle = TESS_DEFAULT_LIGHTSTYLE );

void ChangeMusic( const char* current, const char* fallback, qboolean force );

void ChangeSoundtrack( const char* soundtrack );

class Bubble : public Entity
{
public:
	CLASS_PROTOTYPE( Bubble );

	virtual void Think( Event* ev );
	virtual void Touch( Event* ev );
	virtual void Setup( Vector pos );
};

class Beam : public Entity
{
private:
	Vector start;
	Vector end;

public:
	CLASS_PROTOTYPE( Beam );

	Beam();
	void            setBeam( Vector start, Vector end, int diameter, float r, float g, float b, float alpha, float lifespan );
	virtual void Archive( Archiver& arc );
	virtual void Unarchive( Archiver& arc );
};

inline void Beam::Archive( Archiver& arc )
{
	Entity::Archive( arc );
	arc.WriteVector( start );
	arc.WriteVector( end );
}

inline void Beam::Unarchive( Archiver& arc )
{
	Entity::Unarchive( arc );
	arc.ReadVector( &start );
	arc.ReadVector( &end );
}

class Projectile : public Entity
{
protected:
	int owner;

public:
	CLASS_PROTOTYPE( Projectile );
	float fov;
	Projectile();
	virtual void Setup( Entity* owner, Vector pos, Vector dir );
	virtual void Archive( Archiver& arc );
	virtual void Unarchive( Archiver& arc );
};

inline void Projectile::Archive( Archiver& arc )
{
	Entity::Archive( arc );
	arc.WriteInteger( owner );
	arc.WriteFloat( fov );
}

inline void Projectile::Unarchive( Archiver& arc )
{
	Entity::Unarchive( arc );
	owner = arc.ReadInteger();
	fov = arc.ReadFloat();
}

class FireSprite : public Light
{
public:
	CLASS_PROTOTYPE( FireSprite );

	FireSprite();
	~FireSprite();
};

class FuncBeam : public ScriptSlave
{
protected:
	Entity*      end;
	float damage;
	float life;

public:
	CLASS_PROTOTYPE( FuncBeam );

	FuncBeam();
	void        Activate( Event* ev );
	void        Deactivate( Event* ev );
	void        SetDiameter( Event* ev );
	void        SetLightstyle( Event* ev );
	void        SetMaxoffset( Event* ev );
	void        SetMinoffset( Event* ev );
	void        SetColor( Event* ev );
	void        SetTarget( Event* ev );
	void        Shoot( Vector start, Vector end, int diameter );
	virtual void Archive( Archiver& arc );
	virtual void Unarchive( Archiver& arc );
};

inline void FuncBeam::Archive( Archiver& arc )
{
	ScriptSlave::Archive( arc );
	arc.WriteObjectPointer( end );
	arc.WriteFloat( damage );
	arc.WriteFloat( life );
}

inline void FuncBeam::Unarchive( Archiver& arc )
{
	ScriptSlave::Unarchive( arc );
	arc.ReadObjectPointer( ( Class** )&end );
	damage = arc.ReadFloat();
	life = arc.ReadFloat();
}

class Sprite : public Trigger
{
public:
	CLASS_PROTOTYPE( Sprite );
	Sprite();

	void        Activate( Event* ev );
	void        Deactivate( Event* ev );
};

#endif /* specialfx.h */
