//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/cmodel.c                                               $
// $Revision:: 4                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 2/5/02 7:39 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// model loading
//

#include "qcommon.h"

typedef struct {
	cplane_t	*plane;
	int			children[2];		// negative numbers are leafs
} cnode_t;

typedef struct {
	cplane_t	*plane;
	csurface_t	*surface;
} cbrushside_t;

typedef struct msurface_s {
	int         surfinfo_number;
} msurface_t;

typedef struct {
	int contents;
	int cluster;
	int area;
	unsigned short firstleafbrush;
	unsigned short numleafbrushes;
	msurface_t	**firstmarksurface;
	int nummarksurfaces;
} cleaf_t;

typedef struct {
	int contents;
	int numsides;
	int firstbrushside;
	int checkcount;		// to avoid repeated testings
} cbrush_t;

typedef struct {
	int	numareaportals;
	int	firstareaportal;
	int	floodnum;			// if two areas have equal floodnums, they are connected
	int	floodvalid;
} carea_t;


static int checkcount;
static int emptyleaf;
static char map_name[MAX_QPATH];
static int numbrushsides;
static cbrushside_t *map_brushsides;
static int numplanes;
static cplane_t	*map_planes;
static int numnodes;
static cnode_t *map_nodes;
static int numleafs = 1;	// allow leaf funcs to be called without a map
static cleaf_t *map_leafs;
static int numleafbrushes;
static unsigned short *map_leafbrushes;
static int num_msurfaces;
static msurface_t *map_msurfaces;
static int num_markedsurfaces;
static msurface_t **map_markedsurfaces;
static int numcmodels;
static cmodel_t	*map_cmodels;
static int numbrushes;
static cbrush_t *map_brushes;
static int numvisibility;
static byte *map_visibility;
static dvis_t *map_vis;
static int numentitychars;
static char *map_entitystring;
static int numareas = 1;
static carea_t *map_areas;
static int numareaportals;
static dareaportal_t *map_areaportals;
static qboolean *portalopen;

int map_sequence_number;
int	numtexinfo;
csurface_t *map_surfaces;


static int numclusters = 1;
static csurface_t nullsurface;

static int floodvalid;

static cvar_t *map_noareas;

static void	CM_InitBoxHull (void);
static void	FloodAreaConnections (void);

int		c_pointcontents;
int		c_traces, c_brush_traces;

/*
===============================================================================

					MAP LOADING

===============================================================================
*/

byte	*cmod_base;

static void CMod_Release(void)
{
	Z_Free( map_brushsides );
	numbrushsides = 0;
	map_brushsides = 0;
	Z_Free(map_planes );
	numplanes = 0;
	map_planes = 0;
	Z_Free(map_nodes);
	numnodes = 0;
	map_nodes = 0;
	Z_Free(map_leafs);
	numleafs = 1;	// allow leaf funcs to be called without a map
	map_leafs = 0;
	Z_Free(map_leafbrushes);
	numleafbrushes = 0;
	map_leafbrushes = 0;
	Z_Free(map_msurfaces );
	num_msurfaces = 0;
	map_msurfaces = 0;
	Z_Free(map_markedsurfaces );
	num_markedsurfaces = 0;
	map_markedsurfaces = 0;
	Z_Free( map_cmodels );
	numcmodels = 0;
	map_cmodels = 0;
	Z_Free( map_brushes );
	map_brushes = 0;
	numbrushes = 0;
	Z_Free( map_visibility );
	map_visibility = 0;
	map_vis = ( dvis_t * )0;
	numvisibility = 0;
	Z_Free( map_entitystring );
	map_entitystring = NULL;
	numentitychars = 0;
	Z_Free( map_areas );
	map_areas = 0;
	numareas = 1;
	Z_Free( map_areaportals );
	map_areaportals = 0;
	Z_Free( portalopen );
	portalopen = 0;
	numareaportals = 0;

	Z_Free(map_surfaces);
	numtexinfo = 0;
	map_surfaces = 0;
}

/*
=================
CMod_LoadSubmodels
=================
*/
static void CMod_LoadSubmodels (lump_t *l)
{
	dmodel_t	*in;
	cmodel_t	*out;
	int			i, j, count;

	in = (dmodel_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map with no models");

	if(map_cmodels)
	{
		Z_Free( map_cmodels );
	}

	map_cmodels = NULL;
	numcmodels = count;

	if ( numcmodels )
	{
		map_cmodels = ( cmodel_t * )Z_Malloc(sizeof( cmodel_t ) * numcmodels );
	}

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out = &map_cmodels[i];

		for (j=0 ; j<3 ; j++)
		{	// spread the mins / maxs by a pixel
			out->mins[j] = LittleFloat(in->mins[j]) - 1;
			out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
			out->origin[j] = LittleFloat (in->origin[j]);
		}
		out->headnode = LittleLong (in->headnode);
	}
}

/*
=================
CMod_LoadLightinfos
=================
*/
static void CMod_LoadLightinfos (lump_t *l)
{
	lightvalue_t *in;
	csurface_t *out;
	int i, count, j;
	int style;
	char *stylename;

	in = (lightvalue_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadLightinfos: funny lump size");
	count = l->filelen / sizeof(*in);
	if (count < 1)
		Com_Error (ERR_DROP, "Map with no surfaces");

	if ( !numtexinfo )
		return;

	for ( i=0 ; i<count ; i++, in++)
	{
		style = LittleFloat(in->directstyle );
		stylename = in->directstylename;
		out = map_surfaces;
		for ( j=0 ; j<numtexinfo ; j++, out++)
		{
			if ( !Q_stricmp( stylename, out->groupname ) )
			{
				out->style = style;
			}
		}
	}
}

/*
===================
CMod_ReLoadSurfaces - Reload the surfaces in place to reset their states.
===================
*/
static void CMod_ReLoadSurfaces (lump_t *l)
{
	texinfo_t *in;
	csurface_t *out,*step;
	int count,i,j;
	int next;

	in = (texinfo_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);
	if (count < 1)
		Com_Error (ERR_DROP, "Map with no surfaces");

	if ( numtexinfo == count )
	{
		out = map_surfaces;
		for ( i=0 ; i<count ; i++, in++, out++)
		{
			if (strlen(in->groupname))
				sprintf (out->groupname, "%s", in->groupname);
			else
				sprintf (out->groupname, "%d", i);
      
			out->groupnumber		= i;
			out->trans_mag			= LittleFloat(in->trans_mag);
			out->trans_angle		= LittleLong(in->trans_angle);
			out->translucence		= LittleFloat(in->translucence);
			out->magnitude			= LittleFloat(in->friction);
			out->frequency			= LittleFloat(in->restitution);
			out->flags				= LittleLong(in->flags);
			out->animtime			= LittleFloat(in->animtime);
			out->nonlit				= LittleFloat(in->nonlit);
			out->translucence		= LittleFloat(in->translucence);
			out->animation_frame	= 0;

			for (j=0 ; j<3 ; j++)
				out->color[j] = LittleFloat (in->color[j]);
      
			for (j=0 ; j<3 ; j++)
				out->transvec[j] = -LittleFloat (in->vecs[1][j]);
      
			next = LittleLong (in->nexttexinfo);
			if (next > 0)
			{
				out->next = map_surfaces + next;
			}
			else
			{
				out->next = NULL;
			}
		}      

		// count animation frames
		for (i=0 ; i<count ; i++)
		{
			csurface_t *prev=0;
			out = &map_surfaces[i];
			out->numframes = 1;

			for ( step = out->next ; (step) && (step != out) && (prev != step); prev=step, step=step->next )
			{
				out->numframes++;
			}
			out->frameoffset = rand() % out->numframes;
		}
	}
}

/*
=================
CMod_LoadSurfaces
=================
*/
static void CMod_LoadSurfaces (lump_t *l)
{
	texinfo_t *in;
	csurface_t *out,*step;
	int count,i,j;
	int next;

	in = (texinfo_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);
	if (count < 1)
		Com_Error (ERR_DROP, "Map with no surfaces");

	if(map_surfaces)
	{
		Z_Free(map_surfaces);
	}
	map_surfaces = NULL;
	
	numtexinfo = count;
	
	if ( numtexinfo )
	{
		map_surfaces = ( csurface_t * )Z_Malloc( sizeof( csurface_t ) * numtexinfo );
	}

	out = map_surfaces;
	for ( i=0 ; i<count ; i++, in++, out++)
	{
		if (strlen(in->groupname))
			sprintf (out->groupname, "%s", in->groupname);
		else
			sprintf (out->groupname, "%d", i);
		
		out->groupnumber	= i;
		out->trans_mag		= LittleFloat(in->trans_mag);
		out->trans_angle	= LittleLong(in->trans_angle);
		out->translucence	= LittleFloat(in->translucence);
		out->magnitude		= LittleFloat(in->friction);
		out->frequency		= LittleFloat(in->restitution);
		out->flags			= LittleLong(in->flags);
		out->style			= -1;
		out->animtime		= LittleFloat(in->animtime);
		out->nonlit			= LittleFloat(in->nonlit);
		out->translucence	= LittleFloat(in->translucence);
		
		for (j=0 ; j<3 ; j++)
			out->color[j] = LittleFloat (in->color[j]);
		
		for (j=0 ; j<3 ; j++)
			out->transvec[j] = -LittleFloat (in->vecs[1][j]);
		
		next = LittleLong (in->nexttexinfo);
		if (next > 0)
		{
			out->next = map_surfaces + next;
		}
		else
		{
			out->next = NULL;
		}
	}
		
	// count animation frames
	for (i=0 ; i<count ; i++)
	{
		csurface_t *prev=0;
		out = &map_surfaces[i];
		out->numframes = 1;
		
		for ( step = out->next ; (step) && (step != out) && (prev != step); prev=step, step=step->next )
		{
			out->numframes++;
		}
		out->frameoffset = rand() % out->numframes;
	}
}

/*
=================
CMod_LoadNodes

=================
*/
static void CMod_LoadNodes (lump_t *l)
{
	dnode_t		*in;
	int			child;
	cnode_t		*out;
	int			i, j, count;

	in = (dnode_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map has no nodes");
	if(map_nodes)
	{
		Z_Free( map_nodes );
	}
	map_nodes = NULL;

	numnodes = count;
	if ( numnodes ) {
		map_nodes = ( cnode_t * )Z_Malloc( sizeof( cnode_t ) * ( numnodes + 6 ) );
	}
	out = map_nodes;
		
	for (i=0 ; i<count ; i++) {
		out->plane = map_planes + LittleLong(in->planenum);
		j = 0;
		do {
			child = LittleLong(in->children[j]);
			out->children[j] = child;
		} while (++j<2);
		++out;
		++in;
	}
	
}

/*
=================
CMod_LoadBrushes

=================
*/
static void CMod_LoadBrushes (lump_t *l)
{
	dbrush_t	*in;
	cbrush_t	*out;
	int			i, count;

	in = (dbrush_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if(map_brushes)
	{
		Z_Free( map_brushes );
	}
	map_brushes = NULL;
	
	numbrushes = count;
	if ( numbrushes ) {
		map_brushes = ( cbrush_t * )Z_Malloc( sizeof( cbrush_t ) * ( numbrushes + 1 ) );
	}
	
	out = map_brushes;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		out->firstbrushside = LittleLong(in->firstside);
		out->numsides = LittleLong(in->numsides);
		out->contents = LittleLong(in->contents);

		// patch up fences
		if (out->contents & CONTENTS_DUMMYFENCE)
		{
			out->contents &= ~CONTENTS_WINDOW;
			out->contents &= ~CONTENTS_DUMMYFENCE;
			out->contents |= CONTENTS_FENCE;
		}
	}
}

/*
==============
CMod_LoadFaces
==============
*/
static void CMod_LoadFaces (lump_t *l)
{
	dface_t		*in;
	msurface_t 	*out;
	int			count, msurfnum;

	in = (dface_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "CMod_LoadFaces: funny lump size");
	count = l->filelen / sizeof(*in);

	if(map_msurfaces)
	{
		Z_Free(map_msurfaces );
	}

	map_msurfaces = NULL;
	num_msurfaces = count;
	
	if ( num_msurfaces ) {
		map_msurfaces = ( msurface_t * )Z_Malloc( sizeof( msurface_t ) * ( num_msurfaces + 1 ) );
	}
	
	out = map_msurfaces;
	
	for ( msurfnum=0 ; msurfnum<count ; msurfnum++, in++, out++) {
		out->surfinfo_number = LittleShort (in->texinfo);
	}
}


/*
=======================
CMod_LoadMarkedSurfaces
=======================
*/
static void CMod_LoadMarkedSurfaces (lump_t *l)
{
	short *in;
	msurface_t **out;
	int count, i;

	in = (short *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "CMod_LoadMarkedSurfaces: funny lump size");
	count = l->filelen / sizeof(*in);

	if(map_markedsurfaces)
	{
		Z_Free( map_markedsurfaces );
	}

	map_markedsurfaces = NULL;
	num_markedsurfaces = count;

	if ( num_markedsurfaces ) {
		map_markedsurfaces = ( msurface_t ** )Z_Malloc( sizeof( msurface_t * ) * ( num_markedsurfaces + 1 ) );
	}

	out = map_markedsurfaces;

	for ( i=0 ; i<count ; i++) {
		int j;
		j = LittleShort( in[ i ] );
	
		if (j < 0 ||  j >= num_msurfaces)
			Com_Error (ERR_DROP, "CMod_LoadMarkedSurfaces: bad surface number");
		out[i] = &map_msurfaces[ j ];
	}
}

/*
=================
CMod_LoadLeafs
=================
*/
static void CMod_LoadLeafs (lump_t *l)
{
	int			i;
	cleaf_t		*out;
	dleaf_t 	*in;
	int			count;
	
	in = (dleaf_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);
	
	if (count < 1)
		Com_Error (ERR_DROP, "Map with no leafs");

	if(map_leafs)
	{
		Z_Free( map_leafs );
	}

	map_leafs = NULL;
	
	numleafs = count;

	if ( numleafs ) {
		map_leafs = ( cleaf_t * )Z_Malloc( sizeof( cleaf_t ) * ( numleafs + 1 ) );
	}

	out = map_leafs;
	numclusters = 0;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->contents = LittleLong (in->contents);

		// patch up fences
		if (out->contents & CONTENTS_DUMMYFENCE)
		{
			out->contents &= ~CONTENTS_WINDOW;
			out->contents &= ~CONTENTS_DUMMYFENCE;
			out->contents |= CONTENTS_FENCE;
		}

		out->cluster = LittleShort (in->cluster);
		out->area = LittleShort (in->area);
		out->firstleafbrush = LittleShort (in->firstleafbrush);
		out->numleafbrushes = LittleShort (in->numleafbrushes);


		out->firstmarksurface = &map_markedsurfaces[ LittleUnsignedShort(in->firstleafface) ];
		out->nummarksurfaces = LittleUnsignedShort(in->numleaffaces);

		if (out->cluster >= numclusters)
			numclusters = out->cluster + 1;
	}

	if (map_leafs[0].contents != CONTENTS_SOLID)
		Com_Error (ERR_DROP, "Map leaf 0 is not CONTENTS_SOLID");
	emptyleaf = -1;
	for (i=1 ; i<numleafs ; i++)
	{
		if (!map_leafs[i].contents)
		{
			emptyleaf = i;
			break;
		}
	}
	if (emptyleaf == -1)
		Com_Error (ERR_DROP, "Map does not have an empty leaf");
}

/*
=================
CMod_LoadPlanes
=================
*/
static void CMod_LoadPlanes (lump_t *l)
{
	int			i, j;
	cplane_t	*out;
	dplane_t 	*in;
	int			count;
	int			bits;

	in = (dplane_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map with no planes");

	if(map_planes)
	{
		Z_Free(map_planes );
	}

	numplanes = count;
	if ( numplanes ) {
		map_planes = ( cplane_t * )Z_Malloc( sizeof( cplane_t ) * ( numplanes + 12 ) );
	}
	out = map_planes;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		bits = 0;
		for (j=0 ; j<3 ; j++)
		{
			out->normal[j] = LittleFloat (in->normal[j]);
			if (out->normal[j] < 0)
				bits |= 1<<j;
		}

		out->dist = LittleFloat (in->dist);
		out->type = LittleLong (in->type);
		out->signbits = bits;
	}
}

/*
=================
CMod_LoadLeafBrushes
=================
*/
static void CMod_LoadLeafBrushes (lump_t *l)
{
	int			i;
	unsigned short	*out;
	unsigned short 	*in;
	int			count;

	in = (unsigned short *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if (count < 1)
		Com_Error (ERR_DROP, "Map with no planes");

	if(map_leafbrushes)
	{
		Z_Free(map_leafbrushes);
	}

	map_leafbrushes = 0;
	numleafbrushes = count;
	if ( numleafbrushes ) {
		map_leafbrushes = (unsigned short * )Z_Malloc(sizeof(unsigned short ) * ( numleafbrushes + 1 ) );
	}
	out = map_leafbrushes;
	
	for ( i=0 ; i<count ; i++, in++, out++)
		*out = LittleLong(*in);
}

/*
=================
CMod_LoadBrushSides
=================
*/
static void CMod_LoadBrushSides (lump_t *l)
{
	int			i, j;
	cbrushside_t	*out;
	dbrushside_t 	*in;
	int			count;
	int			num;

	in = (dbrushside_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if(map_brushsides)
	{
		Z_Free(map_brushsides);
	}

	map_brushsides = NULL;
	numbrushsides = count;
	if ( numbrushsides ) {
		map_brushsides = ( cbrushside_t * )Z_Malloc( sizeof( cbrushside_t ) * ( numbrushsides + 6 ) );
	}
	out = map_brushsides;

	for ( i=0 ; i<count ; i++, in++, out++) {
		num = LittleShort (in->planenum);
		out->plane = &map_planes[num];
		j = LittleShort (in->texinfo);
		if (j >= numtexinfo)
			Com_Error (ERR_DROP, "Bad brushside texinfo");
		if ( j < 0 )
			out->surface = &nullsurface;
		else
			out->surface = &map_surfaces[j];
	}
}

/*
=================
CMod_LoadAreas
=================
*/
static void CMod_LoadAreas (lump_t *l)
{
	int			i;
	carea_t		*out;
	darea_t 	*in;
	int			count;

	in = (darea_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if(map_areas)
	{
		Z_Free( map_areas );
	}

	map_areas = 0;
	
	numareas = count;
	if ( numareas ) {
		map_areas = ( carea_t * )Z_Malloc( sizeof( carea_t ) * numareas );
	}
	out = map_areas;
	
	for ( i=0 ; i<count ; i++, in++, out++) {
		out->numareaportals = LittleLong (in->numareaportals);
		out->firstareaportal = LittleLong (in->firstareaportal);
		out->floodvalid = 0;
		out->floodnum = 0;
	}
}

/*
=================
CMod_LoadAreaPortals
=================
*/
static void CMod_LoadAreaPortals (lump_t *l)
{
	int			i;
	dareaportal_t		*out;
	dareaportal_t 	*in;
	int			count;

	in = (dareaportal_t *)(cmod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Com_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size");
	count = l->filelen / sizeof(*in);

	if(map_areaportals)
	{
		Z_Free( map_areaportals );
	}

	map_areaportals = NULL;

	if(portalopen)
	{
		Z_Free( portalopen );
	}

	portalopen = NULL;
	numareaportals = count;
	if ( numareaportals ) {
		map_areaportals = ( dareaportal_t * )Z_Malloc( sizeof( dareaportal_t ) * numareaportals );
		portalopen = ( qboolean * )Z_Malloc( sizeof( qboolean ) * numareaportals );
	}
	out = map_areaportals;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->portalnum = LittleLong (in->portalnum);
		out->otherarea = LittleLong (in->otherarea);
	}
}

/*
=================
CMod_LoadVisibility
=================
*/
static void CMod_LoadVisibility (lump_t *l)
{
	int		i;
	
	numvisibility = l->filelen;

	if(map_visibility)
	{
		Z_Free( map_visibility );
	}

	map_visibility = 0;
	map_vis = ( dvis_t * )0;

	if ( numvisibility + sizeof( dvis_t ) ) {
		map_visibility = ( byte * )Z_Malloc( sizeof( byte ) * (numvisibility + sizeof( dvis_t ) ) );
		map_vis = ( dvis_t * )map_visibility;
	}
	
	if ( l->filelen ) {
		memcpy(map_visibility, cmod_base + l->fileofs, l->filelen);
	}	
	map_vis->numclusters = LittleLong (map_vis->numclusters);
	for (i=0 ; i<map_vis->numclusters ; i++) {
		map_vis->bitofs[i][0] = LittleLong (map_vis->bitofs[i][0]);
		map_vis->bitofs[i][1] = LittleLong (map_vis->bitofs[i][1]);
	}
}


/*
=================
CMod_LoadEntityString
=================
*/
static void CMod_LoadEntityString (lump_t *l)
{
	numentitychars = l->filelen;

	if(map_entitystring)
	{
		Z_Free( map_entitystring );
	}

	map_entitystring = NULL;
	if ( numentitychars ) {
		map_entitystring = ( char * )Z_Malloc( sizeof( char ) * numentitychars );
	}
	
	memcpy(map_entitystring, cmod_base + l->fileofs, l->filelen);
}

/*
=============================
CM_MarkLeafSurfacesAndCleanup
=============================
*/
static void CMod_MarkLeafSurfacesAndCleanup ( void )
{
	int i,c;
	cleaf_t *leaf;
	msurface_t **mark;

	// Go through all the leafs in the world
	// and set the surfinfos of the marked surfaces to 
	// the correct leaf.
	for ( i=0; i<numleafs; i++ )
	{
		leaf = &map_leafs[ i ];

		mark = leaf->firstmarksurface;
		c    = leaf->nummarksurfaces;

		if (c)
		{  
			do
			{
				map_surfaces[(*mark)->surfinfo_number].leaf = i;
				mark++;
			} while (--c);
		}
	}

	// Z_Free up the data since we don't need it anymore
	if(map_markedsurfaces)
	{
		Z_Free(map_markedsurfaces);
	}

	map_markedsurfaces = 0;
	num_markedsurfaces = 0;

	if(map_msurfaces)
	{
		Z_Free(map_msurfaces);
	}

	map_msurfaces = 0;
	num_msurfaces = 0;
}

/*
==================
CM_LoadMap

Loads in the map and all submodels
==================
*/
cmodel_t *CM_LoadMap (const char *name, qboolean clientload, unsigned *checksum)
{
	void		*buf;
	int				i;
	dheader_t		header;
	int				length;
	static unsigned	last_checksum;

	memset( &nullsurface, 0, sizeof( csurface_t ) );
	nullsurface.magnitude = 1.0f;

	map_noareas = Cvar_Get ("map_noareas", "0", 0);

	if (  map_name[0] && !strcmp (map_name, name) && (clientload || !Cvar_VariableValue ("flushmap")) )
	{
		*checksum = last_checksum;
		if (!clientload)
		{
			memset( portalopen, 0, numareaportals * sizeof(qboolean) );
			FloodAreaConnections ();
		}

		// Don't force the renderer to reload the map
		Cvar_Set( "forcereload", "0" );

		if ( clientload )
		{
			// Reload the surfaces in place on the client before returning
			length = FS_LoadFile (name, &buf);
			if (!buf)
				Com_Error (ERR_DROP, "Couldn't load %s", name);

			header = *(dheader_t *)buf;
			for (i=0 ; i<sizeof(dheader_t)/4 ; i++)
				((int *)&header)[i] = LittleLong ( ((int *)&header)[i]);

			if (header.version != BSPVERSION)
				Com_Error (ERR_DROP, "CMod_LoadBrushModel: %s has wrong version number (%i should be %i)", name, header.version, BSPVERSION);
   
			cmod_base = (byte *)buf;

			CMod_ReLoadSurfaces (&header.lumps[LUMP_TEXINFO]);
			FS_FreeFile( buf );
		}

		return &map_cmodels[0];		// still have the right version
	}

	// free old stuff
	numplanes = 0;
	numnodes = 0;
	numleafs = 0;
	numcmodels = 0;
	numvisibility = 0;
	numentitychars = 0;
	if(map_entitystring)
	{
		Z_Free( map_entitystring );
	}
	map_entitystring = NULL;
	map_name[0] = 0;

	if (!name || !name[0])
	{
		numleafs = 1;
		numclusters = 1;
		numareas = 1;
		*checksum = 0;

		if (!map_cmodels) {
			map_cmodels = ( cmodel_t * )Z_Malloc( sizeof( cmodel_t ));
		}
		return &map_cmodels[0];			// cinematic servers won't have anything at all
	}

	//
	// load the file
	//
	length = FS_LoadFile (name, &buf);
	if (!buf)
		Com_Error (ERR_DROP, "Couldn't load %s", name);

	last_checksum = LittleLong( Com_BlockChecksum (buf, length) );
	*checksum = last_checksum;

	header = *(dheader_t *)buf;
	for (i=0 ; i<sizeof(dheader_t)/4 ; i++)
		((int *)&header)[i] = LittleLong ( ((int *)&header)[i]);

	if (header.version != BSPVERSION)
		Com_Error (ERR_DROP, "CMod_LoadBrushModel: %s has wrong version number (%i should be %i)"
		, name, header.version, BSPVERSION);

	cmod_base = (byte *)buf;

	// load into heap
	CMod_LoadSurfaces (&header.lumps[LUMP_TEXINFO]);
	CMod_LoadLightinfos (&header.lumps[LUMP_LIGHTINFO]);
	CMod_LoadFaces (&header.lumps[LUMP_FACES]);
	CMod_LoadMarkedSurfaces (&header.lumps[LUMP_LEAFFACES]);
	CMod_LoadLeafs (&header.lumps[LUMP_LEAFS]);
	CMod_LoadLeafBrushes (&header.lumps[LUMP_LEAFBRUSHES]);
	CMod_LoadPlanes (&header.lumps[LUMP_PLANES]);
	CMod_LoadBrushes (&header.lumps[LUMP_BRUSHES]);
	CMod_LoadBrushSides (&header.lumps[LUMP_BRUSHSIDES]);
	CMod_LoadSubmodels (&header.lumps[LUMP_MODELS]);
	CMod_LoadNodes (&header.lumps[LUMP_NODES]);
	CMod_LoadAreas (&header.lumps[LUMP_AREAS]);
	CMod_LoadAreaPortals (&header.lumps[LUMP_AREAPORTALS]);
	CMod_LoadVisibility (&header.lumps[LUMP_VISIBILITY]);
	CMod_LoadEntityString (&header.lumps[LUMP_ENTITIES]);
	CMod_MarkLeafSurfacesAndCleanup();

	FS_FreeFile (buf);

	CM_InitBoxHull ();

	memset( portalopen, 0, numareaportals * sizeof( qboolean ) );

	FloodAreaConnections ();

	strcpy (map_name, name);

	map_sequence_number++;
	Com_DPrintf(DEVELOPER_MSG_STANDARD, "Updating map sequence number to %i\n", map_sequence_number );
   
	// Force the renderer to reload since we updated the surfaces
	Cvar_Set( "forcereload", "1" );

	return &map_cmodels[0];
}

/*
==================
CM_InlineModel
==================
*/
cmodel_t	*CM_InlineModel (const char *name)
{
	int		num;

	if (!name || name[0] != '*')
		Com_Error (ERR_DROP, "CM_InlineModel: bad name");
	num = atoi (name+1);
	if (num < 1 || num >= numcmodels)
		Com_Error (ERR_DROP, "CM_InlineModel: bad number");

	return &map_cmodels[num];
}

int		CM_NumClusters (void)
{
	return numclusters;
}

int		CM_NumInlineModels (void)
{
	return numcmodels;
}

char	*CM_EntityString (void)
{
	return map_entitystring;
}

int		CM_LeafContents (int leafnum)
{
	if (leafnum < 0 || leafnum >= numleafs)
		Com_Error (ERR_DROP, "CM_LeafContents: bad number");
	return map_leafs[leafnum].contents;
}

int		CM_LeafCluster (int leafnum)
{
	if (leafnum < 0 || leafnum >= numleafs)
		Com_Error (ERR_DROP, "CM_LeafCluster: bad number");
	return map_leafs[leafnum].cluster;
}

int		CM_LeafArea (int leafnum)
{
	if (leafnum < 0 || leafnum >= numleafs)
		Com_Error (ERR_DROP, "CM_LeafArea: bad number");
	return map_leafs[leafnum].area;
}

csurface_t *CM_Surfinfos (void)
{
	return &map_surfaces[0];
}

int CM_NumSurfinfos (void)
{
	return numtexinfo;
}


//=======================================================================


cplane_t	*box_planes;
int			box_headnode;
cbrush_t	*box_brush;
cleaf_t		*box_leaf;

/*
===================
CM_InitBoxHull

Set up the planes and nodes so that the six floats of a bounding box
can just be stored out and get a proper clipping hull structure.
===================
*/
static void CM_InitBoxHull (void)
{
	int			i;
	int			side;
	cnode_t		*c;
	cplane_t	*p;
	cbrushside_t	*s;

	box_headnode = numnodes;
	box_planes = &map_planes[numplanes];

	box_brush = &map_brushes[numbrushes];
	box_brush->numsides = 6;
	box_brush->firstbrushside = numbrushsides;
	box_brush->contents = CONTENTS_MONSTER;

	box_leaf = &map_leafs[numleafs];
	box_leaf->contents = CONTENTS_MONSTER;
	box_leaf->firstleafbrush = numleafbrushes;
	box_leaf->numleafbrushes = 1;

	map_leafbrushes[numleafbrushes] = numbrushes;

	for (i=0 ; i<6 ; i++)
	{
		side = i&1;

		// brush sides
		s = &map_brushsides[numbrushsides+i];
		s->plane = 	map_planes + (numplanes+i*2+side);
		s->surface = &nullsurface;

		// nodes
		c = &map_nodes[box_headnode+i];
		c->plane = map_planes + (numplanes+i*2);
		c->children[side] = -1 - emptyleaf;
		if (i != 5)
			c->children[side^1] = box_headnode+i + 1;
		else
			c->children[side^1] = -1 - numleafs;

		// planes
		p = &box_planes[i*2];
		p->type = i>>1;
		p->signbits = 0;
		VectorClear (p->normal);
		p->normal[i>>1] = 1;

		p = &box_planes[i*2+1];
		p->type = 3 + (i>>1);
		p->signbits = 0;
		VectorClear (p->normal);
		p->normal[i>>1] = -1;
	}
}


/*
===================
CM_HeadnodeForBox

To keep everything totally uniform, bounding boxes are turned into small
BSP trees instead of being compared directly.
===================
*/
int	CM_HeadnodeForBox (vec3_t mins, vec3_t maxs)
{
	box_planes[0].dist = maxs[0];
	box_planes[1].dist = -maxs[0];
	box_planes[2].dist = mins[0];
	box_planes[3].dist = -mins[0];
	box_planes[4].dist = maxs[1];
	box_planes[5].dist = -maxs[1];
	box_planes[6].dist = mins[1];
	box_planes[7].dist = -mins[1];
	box_planes[8].dist = maxs[2];
	box_planes[9].dist = -maxs[2];
	box_planes[10].dist = mins[2];
	box_planes[11].dist = -mins[2];

	return box_headnode;
}


/*
==================
CM_PointLeafnum_r

==================
*/
static int CM_PointLeafnum_r (const vec3_t p, int num)
{
	float		d;
	cnode_t		*node;
	cplane_t	*plane;

	while (num >= 0)
	{
		node = map_nodes + num;
		plane = node->plane;

		if (plane->type < 3)
			d = p[plane->type] - plane->dist;
		else
			d = DotProduct (plane->normal, p) - plane->dist;
		if (d < 0)
			num = node->children[1];
		else
			num = node->children[0];
	}

	c_pointcontents++;		// optimize counter

	return -1 - num;
}

int CM_PointLeafnum (const vec3_t p)
{
	if (!numplanes)
		return 0;		// sound may call this without map loaded
	return CM_PointLeafnum_r (p, 0);
}



/*
=============
CM_BoxLeafnums

Fills in a list of all the leafs touched
=============
*/
int		leaf_count, leaf_maxcount;
int		*leaf_list;
float	*leaf_mins, *leaf_maxs;
int		leaf_topnode;

static void CM_BoxLeafnums_r (int nodenum)
{
	cplane_t	*plane;
	cnode_t		*node;
	int		s;

	while (1)
	{
		if (nodenum < 0)
		{
			if (leaf_count >= leaf_maxcount)
			{
//				Com_Printf ("CM_BoxLeafnums_r: overflow\n");
				return;
			}
			leaf_list[leaf_count++] = -1 - nodenum;
			return;
		}

		node = &map_nodes[nodenum];
		plane = node->plane;
      //s = BoxOnPlaneSide (leaf_mins, leaf_maxs, plane);
      s = BOX_ON_PLANE_SIDE(leaf_mins, leaf_maxs, plane);

      if (s == 1)
			nodenum = node->children[0];
		else if (s == 2)
			nodenum = node->children[1];
		else
		{	// go down both
			if (leaf_topnode == -1)
				leaf_topnode = nodenum;
			CM_BoxLeafnums_r (node->children[0]);
			nodenum = node->children[1];
		}

	}
}

static int	CM_BoxLeafnums_headnode (vec3_t mins, vec3_t maxs, int *list, int listsize, int headnode, int *topnode)
{
	leaf_list = list;
	leaf_count = 0;
	leaf_maxcount = listsize;
	leaf_mins = mins;
	leaf_maxs = maxs;

	leaf_topnode = -1;

	CM_BoxLeafnums_r (headnode);

	if (topnode)
		*topnode = leaf_topnode;

	return leaf_count;
}

int	CM_BoxLeafnums (vec3_t mins, vec3_t maxs, int *list, int listsize, int *topnode)
{
	return CM_BoxLeafnums_headnode (mins, maxs, list,
		listsize, map_cmodels[0].headnode, topnode);
}



/*
==================
CM_PointContents

==================
*/
int CM_PointContents (vec3_t p, int headnode)
{
	int		l;

	if (!numnodes)	// map not loaded
		return 0;

	l = CM_PointLeafnum_r (p, headnode);

	return map_leafs[l].contents;
}

/*
==================
CM_TransformedPointContents

Handles offseting and rotation of the end points for moving and
rotating entities
==================
*/
int	CM_TransformedPointContents (vec3_t p, int headnode, vec3_t origin, vec3_t angles)
{
	vec3_t		p_l;
	vec3_t		temp;
	vec3_t		forward, right, up;
	int			l;

	// subtract origin offset
	VectorSubtract (p, origin, p_l);

	// rotate start and end into the models frame of reference
	if (headnode != box_headnode &&
	(angles[0] || angles[1] || angles[2]) )
	{
		AngleVectors (angles, forward, right, up);

		VectorCopy (p_l, temp);
		p_l[0] = DotProduct (temp, forward);
		p_l[1] = -DotProduct (temp, right);
		p_l[2] = DotProduct (temp, up);
	}

	l = CM_PointLeafnum_r (p_l, headnode);

	return map_leafs[l].contents;
}


/*
===============================================================================

BOX TRACING

===============================================================================
*/

// 1/32 epsilon to keep floating point happy
#define	DIST_EPSILON	(0.03125)

vec3_t	trace_start, trace_end;
vec3_t	trace_mins, trace_maxs;
vec3_t	trace_extents;

trace_t	trace_trace;
int		trace_contents;
qboolean	trace_ispoint;		// optimized case

/*
================
CM_ClipBoxToBrush
================
*/
static void CM_ClipBoxToBrush (vec3_t mins, vec3_t maxs, vec3_t p1, vec3_t p2,
					  trace_t *trace, cbrush_t *brush)
{
	int			i, j;
	cplane_t	*plane, *clipplane;
	float		dist;
	float		enterfrac, leavefrac;
	vec3_t		ofs;
	float		d1, d2;
	qboolean	getout, startout;
	float		f;
	cbrushside_t	*side, *leadside;

	enterfrac = -1;
	leavefrac = 1;
	clipplane = NULL;

	if (!brush->numsides)
		return;

	c_brush_traces++;

	getout = false;
	startout = false;
	leadside = NULL;

	for (i=0 ; i<brush->numsides ; i++)
	{
		side = &map_brushsides[brush->firstbrushside+i];
		plane = side->plane;

		// FIXME: special case for axial

		if (!trace_ispoint)
		{	// general box case

			// push the plane out apropriately for mins/maxs

			// FIXME: use signbits into 8 way lookup for each mins/maxs
			for (j=0 ; j<3 ; j++)
			{
				if (plane->normal[j] < 0)
					ofs[j] = maxs[j];
				else
					ofs[j] = mins[j];
			}
			dist = DotProduct (ofs, plane->normal);
			dist = plane->dist - dist;
		}
		else
		{	// special point case
			dist = plane->dist;
		}

		d1 = DotProduct (p1, plane->normal) - dist;
		d2 = DotProduct (p2, plane->normal) - dist;

		if (d2 > 0)
			getout = true;	// endpoint is not in solid
		if (d1 > 0)
			startout = true;

		// if completely in front of face, no intersection
		if (d1 > 0 && d2 >= d1)
			return;

		if (d1 <= 0 && d2 <= 0)
			continue;

		// crosses face
		if (d1 > d2)
		{	// enter
			f = (d1-DIST_EPSILON) / (d1-d2);
			if (f > enterfrac)
			{
				enterfrac = f;
				clipplane = plane;
				leadside = side;
			}
		}
		else
		{	// leave
			f = (d1+DIST_EPSILON) / (d1-d2);
			if (f < leavefrac)
				leavefrac = f;
		}
	}

	if (!startout)
	{	// original point was inside brush
		trace->startsolid = true;
		if (!getout)
			trace->allsolid = true;
		return;
	}
	if (enterfrac < leavefrac)
	{
		if (enterfrac > -1 && enterfrac < trace->fraction)
		{
			if (enterfrac < 0)
				enterfrac = 0;
			trace->fraction = enterfrac;
			trace->plane = *clipplane;
			trace->surface = leadside->surface;
			trace->contents = brush->contents;
		}
	}
}

/*
================
CM_TestBoxInBrush
================
*/
static void CM_TestBoxInBrush (vec3_t mins, vec3_t maxs, vec3_t p1,
					  trace_t *trace, cbrush_t *brush)
{
	int			i, j;
	cplane_t	*plane;
	float		dist;
	vec3_t		ofs;
	float		d1;
	cbrushside_t	*side;

	if (!brush->numsides)
		return;

	for (i=0 ; i<brush->numsides ; i++)
	{
		side = &map_brushsides[brush->firstbrushside+i];
		plane = side->plane;

		// FIXME: special case for axial

		// general box case

		// push the plane out apropriately for mins/maxs

		// FIXME: use signbits into 8 way lookup for each mins/maxs
		for (j=0 ; j<3 ; j++)
		{
			if (plane->normal[j] < 0)
				ofs[j] = maxs[j];
			else
				ofs[j] = mins[j];
		}
		dist = DotProduct (ofs, plane->normal);
		dist = plane->dist - dist;

		d1 = DotProduct (p1, plane->normal) - dist;

		// if completely in front of face, no intersection
		if (d1 > 0)
			return;

	}

	// inside this brush
	trace->startsolid = trace->allsolid = true;
	trace->fraction = 0;
	trace->contents = brush->contents;
}


/*
================
CM_TraceToLeaf
================
*/
static void CM_TraceToLeaf (int leafnum)
{
	int			k;
	int			brushnum;
	cleaf_t		*leaf;
	cbrush_t	*b;

	leaf = &map_leafs[leafnum];
	if ( !(leaf->contents & trace_contents))
		return;
	// trace line against all brushes in the leaf
	for (k=0 ; k<leaf->numleafbrushes ; k++)
	{
		brushnum = map_leafbrushes[leaf->firstleafbrush+k];
		b = &map_brushes[brushnum];
		if (b->checkcount == checkcount)
			continue;	// allready checked this brush in another leaf
		b->checkcount = checkcount;

		if ( !(b->contents & trace_contents))
			continue;
		CM_ClipBoxToBrush (trace_mins, trace_maxs, trace_start, trace_end, &trace_trace, b);
		if (!trace_trace.fraction)
			return;
	}

}


/*
================
CM_TestInLeaf
================
*/
static void CM_TestInLeaf (int leafnum)
{
	int			k;
	int			brushnum;
	cleaf_t		*leaf;
	cbrush_t	*b;

	leaf = &map_leafs[leafnum];
	if ( !(leaf->contents & trace_contents))
		return;
	// trace line against all brushes in the leaf
	for (k=0 ; k<leaf->numleafbrushes ; k++)
	{
		brushnum = map_leafbrushes[leaf->firstleafbrush+k];
		b = &map_brushes[brushnum];
		if (b->checkcount == checkcount)
			continue;	// allready checked this brush in another leaf
		b->checkcount = checkcount;

		if ( !(b->contents & trace_contents))
			continue;
		CM_TestBoxInBrush (trace_mins, trace_maxs, trace_start, &trace_trace, b);
		if (!trace_trace.fraction)
			return;
	}

}


/*
==================
CM_RecursiveHullCheck

==================
*/
static void CM_RecursiveHullCheck (int num, float p1f, float p2f,const vec3_t p1,const vec3_t p2)
{
	cnode_t		*node;
	cplane_t	*plane;
	float		t1, t2, offset;
	float		frac, frac2;
	float		idist;
	int			i;
	vec3_t		mid;
	int			side;
	float		midf;

	if (trace_trace.fraction <= p1f)
		return;		// allready hit something nearer

	// if < 0, we are in a leaf node
	if (num < 0)
	{
		CM_TraceToLeaf (-1-num);
		return;
	}

	//
	// find the point distances to the seperating plane
	// and the offset for the size of the box
	//
	node = map_nodes + num;
	plane = node->plane;

	if (plane->type < 3)
	{
		t1 = p1[plane->type] - plane->dist;
		t2 = p2[plane->type] - plane->dist;
		offset = trace_extents[plane->type];
	}
	else
	{
		t1 = DotProduct (plane->normal, p1) - plane->dist;
		t2 = DotProduct (plane->normal, p2) - plane->dist;
		if (trace_ispoint)
			offset = 0;
		else
			offset = fabs(trace_extents[0]*plane->normal[0]) +
				fabs(trace_extents[1]*plane->normal[1]) +
				fabs(trace_extents[2]*plane->normal[2]);
	}


#if 0
CM_RecursiveHullCheck (node->children[0], p1f, p2f, p1, p2);
CM_RecursiveHullCheck (node->children[1], p1f, p2f, p1, p2);
return;
#endif

	// see which sides we need to consider
	if (t1 >= offset && t2 >= offset)
	{
		CM_RecursiveHullCheck (node->children[0], p1f, p2f, p1, p2);
		return;
	}
	if (t1 < -offset && t2 < -offset)
	{
		CM_RecursiveHullCheck (node->children[1], p1f, p2f, p1, p2);
		return;
	}

	// put the crosspoint DIST_EPSILON pixels on the near side
	if (t1 < t2)
	{
		idist = 1.0/(t1-t2);
		side = 1;
		frac2 = (t1 + offset + DIST_EPSILON)*idist;
		frac = (t1 - offset + DIST_EPSILON)*idist;
	}
	else if (t1 > t2)
	{
		idist = 1.0/(t1-t2);
		side = 0;
		frac2 = (t1 - offset - DIST_EPSILON)*idist;
		frac = (t1 + offset + DIST_EPSILON)*idist;
	}
	else
	{
		side = 0;
		frac = 1;
		frac2 = 0;
	}

	// move up to the node
	if (frac < 0)
		frac = 0;
	if (frac > 1)
		frac = 1;

	midf = p1f + (p2f - p1f)*frac;
	for (i=0 ; i<3 ; i++)
		mid[i] = p1[i] + frac*(p2[i] - p1[i]);

	CM_RecursiveHullCheck (node->children[side], p1f, midf, p1, mid);


	// go past the node
	if (frac2 < 0)
		frac2 = 0;
	if (frac2 > 1)
		frac2 = 1;

	midf = p1f + (p2f - p1f)*frac2;
	for (i=0 ; i<3 ; i++)
		mid[i] = p1[i] + frac2*(p2[i] - p1[i]);

	CM_RecursiveHullCheck (node->children[side^1], midf, p2f, mid, p2);
}



//======================================================================

/*
==================
CM_BoxTrace
==================
*/
trace_t		CM_BoxTrace (const vec3_t start,const vec3_t end,
						  const vec3_t mins,const vec3_t maxs,
						  int headnode, int brushmask)
{
	int		i;

	checkcount++;		// for multi-check avoidance

	c_traces++;			// for statistics, may be zeroed

	// fill in a default trace
	memset (&trace_trace, 0, sizeof(trace_trace));
	trace_trace.fraction = 1;
	trace_trace.surface = &nullsurface;

	if (!numnodes)	// map not loaded
		return trace_trace;

	trace_contents = brushmask;
	VectorCopy (start, trace_start);
	VectorCopy (end, trace_end);
	VectorCopy(mins,trace_mins);
	VectorCopy(maxs,trace_maxs);
	VectorSubtract( end, start, trace_trace.dir );
	VectorNormalize( trace_trace.dir );

	//
	// check for position test special case
	//
	if (start[0] == end[0] && start[1] == end[1] && start[2] == end[2])
	{
		int		leafs[1024];
		int		i, numleafs;
		vec3_t	c1, c2;
		int		topnode;

		VectorAdd (start, mins, c1);
		VectorAdd (start, maxs, c2);
		for (i=0 ; i<3 ; i++)
		{
			c1[i] -= 1;
			c2[i] += 1;
		}

		numleafs = CM_BoxLeafnums_headnode (c1, c2, leafs, 1024, headnode, &topnode);
		for (i=0 ; i<numleafs ; i++)
		{
			CM_TestInLeaf (leafs[i]);
			if (trace_trace.allsolid)
				break;
		}
		VectorCopy (start, trace_trace.endpos);
		return trace_trace;
	}

	//
	// check for point special case
	//
	if (mins[0] == 0 && mins[1] == 0 && mins[2] == 0
		&& maxs[0] == 0 && maxs[1] == 0 && maxs[2] == 0)
	{
		trace_ispoint = true;
		VectorClear (trace_extents);
	}
	else
	{
		trace_ispoint = false;
		trace_extents[0] = -mins[0] > maxs[0] ? -mins[0] : maxs[0];
		trace_extents[1] = -mins[1] > maxs[1] ? -mins[1] : maxs[1];
		trace_extents[2] = -mins[2] > maxs[2] ? -mins[2] : maxs[2];
	}

	//
	// general sweeping through world
	//
	CM_RecursiveHullCheck (headnode, 0, 1, start, end);

	if (trace_trace.fraction == 1)
	{
		VectorCopy (end, trace_trace.endpos);
	}
	else
	{
		for (i=0 ; i<3 ; i++)
			trace_trace.endpos[i] = start[i] + trace_trace.fraction * (end[i] - start[i]);
	}
	return trace_trace;
}


/*
==================
CM_TransformedBoxTrace

Handles offseting and rotation of the end points for moving and
rotating entities
==================
*/
#ifdef _MSC_VER
#pragma optimize("", off)
#endif

trace_t		CM_TransformedBoxTrace (const vec3_t start,const vec3_t end,
						  const vec3_t mins,const vec3_t maxs,
						  int headnode, int brushmask,
						  vec3_t origin, const vec3_t angles)
{
	trace_t		trace;
	vec3_t		start_l, end_l;
	vec3_t		a;
	vec3_t		forward, right, up;
	vec3_t		temp;
	qboolean	rotated;

	// subtract origin offset
	VectorSubtract (start, origin, start_l);
	VectorSubtract (end, origin, end_l);

	// rotate start and end into the models frame of reference
	if (headnode != box_headnode &&
		(angles[0] || angles[1] || angles[2]) )
		rotated = true;
	else
		rotated = false;

	if (rotated)
	{
		AngleVectors (angles, forward, right, up);

		VectorCopy (start_l, temp);
		start_l[0] = DotProduct (temp, forward);
		start_l[1] = -DotProduct (temp, right);
		start_l[2] = DotProduct (temp, up);

		VectorCopy (end_l, temp);
		end_l[0] = DotProduct (temp, forward);
		end_l[1] = -DotProduct (temp, right);
		end_l[2] = DotProduct (temp, up);
	}

	// sweep the box through the model
	trace = CM_BoxTrace (start_l, end_l, mins, maxs, headnode, brushmask);

	if (rotated && trace.fraction != 1.0)
	{
		// FIXME: figure out how to do this with existing angles
		VectorNegate (angles, a);
		AngleVectors (a, forward, right, up);

		VectorCopy (trace.plane.normal, temp);
		trace.plane.normal[0] = DotProduct (temp, forward);
		trace.plane.normal[1] = -DotProduct (temp, right);
		trace.plane.normal[2] = DotProduct (temp, up);
	}

	trace.endpos[0] = start[0] + trace.fraction * (end[0] - start[0]);
	trace.endpos[1] = start[1] + trace.fraction * (end[1] - start[1]);
	trace.endpos[2] = start[2] + trace.fraction * (end[2] - start[2]);

	return trace;
}

#ifdef _MSC_VER
#pragma optimize("", on)
#endif


/*
===============================================================================

PVS / PHS

===============================================================================
*/

/*
===================
CM_DecompressVis
===================
*/
static void CM_DecompressVis (byte *in, byte *out)
{
	int		c;
	byte	*out_p;
	int		row;

	row = (numclusters+7)>>3;
	out_p = out;

	if (!in || !numvisibility)
	{	// no vis info, so make all visible
		while (row)
		{
			*out_p++ = 0xff;
			row--;
		}
		return;
	}

	do
	{
		if (*in)
		{
			*out_p++ = *in++;
			continue;
		}

		c = in[1];
		in += 2;
		if ((out_p - out) + c > row)
		{
			c = row - (out_p - out);
			Com_DPrintf (DEVELOPER_MSG_STANDARD, "warning: Vis decompression overrun\n");
		}
		while (c)
		{
			*out_p++ = 0;
			c--;
		}
	} while (out_p - out < row);
}

byte	pvsrow[MAX_MAP_LEAFS/8];
byte	phsrow[MAX_MAP_LEAFS/8];

byte	*CM_ClusterPVS (int cluster)
{
	if ( cluster == -1 )
		memset (pvsrow, 0, (numclusters+7)>>3);
	else if ( !map_vis->numclusters )
		memset (pvsrow, 0xff, (numclusters+7)>>3);
	else
		CM_DecompressVis (map_visibility + map_vis->bitofs[cluster][DVIS_PVS], pvsrow);
	return pvsrow;
}

byte	*CM_ClusterPHS (int cluster)
{
	if ( cluster == -1 )
		memset (phsrow, 0, (numclusters+7)>>3);
	else if ( !map_vis->numclusters )
		memset (phsrow, 0xff, (numclusters+7)>>3);
	else
		CM_DecompressVis (map_visibility + map_vis->bitofs[cluster][DVIS_PHS], phsrow);
	return phsrow;
}


/*
===============================================================================

AREAPORTALS

===============================================================================
*/

static void FloodArea_r (carea_t *area, int floodnum)
{
	int		i;
	dareaportal_t	*p;

	if (area->floodvalid == floodvalid)
	{
		if (area->floodnum == floodnum)
			return;
      Com_DPrintf(DEVELOPER_MSG_STANDARD, "FloodArea_r : reflooded area\n");
	}

	area->floodnum = floodnum;
	area->floodvalid = floodvalid;
	p = &map_areaportals[area->firstareaportal];
	for (i=0 ; i<area->numareaportals ; i++, p++)
	{
		if (portalopen[p->portalnum])
			FloodArea_r (&map_areas[p->otherarea], floodnum);
	}
}

/*
====================
FloodAreaConnections


====================
*/
static void	FloodAreaConnections (void)
{
	int		i;
	carea_t	*area;
	int		floodnum;

	// all current floods are now invalid
	floodvalid++;
	floodnum = 0;

	// area 0 is not used
	for (i=1 ; i<numareas ; i++)
	{
		area = &map_areas[i];
		if (area->floodvalid == floodvalid)
			continue;		// allready flooded into
		floodnum++;
		FloodArea_r (area, floodnum);
	}

}

void	CM_SetAreaPortalState (int portalnum, qboolean open)
{
	if (portalnum > numareaportals)
		Com_Error (ERR_DROP, "areaportal > numareaportals");

	portalopen[portalnum] = open;
	FloodAreaConnections ();
}

qboolean	CM_AreasConnected (int area1, int area2)
{
	if (map_noareas->value)
		return true;

	if (area1 > numareas || area2 > numareas)
		Com_Error (ERR_DROP, "area > numareas");

	if (map_areas[area1].floodnum == map_areas[area2].floodnum)
		return true;
	return false;
}


/*
=================
CM_WriteAreaBits

Writes a length byte followed by a bit vector of all the areas
that area in the same flood as the area parameter

This is used by the client refreshes to cull visibility
=================
*/
int CM_WriteAreaBits (byte *buffer, int area)
{
	int		i;
	int		floodnum;
	int		bytes;

	bytes = (numareas+7)>>3;

	if (map_noareas->value)
	{	// for debugging, send everything
		memset (buffer, 255, bytes);
	}
	else
	{
		memset (buffer, 0, bytes);

		floodnum = map_areas[area].floodnum;
		for (i=0 ; i<numareas ; i++)
		{
			if (map_areas[i].floodnum == floodnum || !area)
				buffer[i>>3] |= 1<<(i&7);
		}
	}

	return bytes;
}


/*
===================
CM_WritePortalState

Writes the portal state to a savegame file
===================
*/
void CM_WritePortalState (FILE *f)
{
	fwrite (&numareaportals, sizeof(numareaportals), 1, f);
	fwrite (portalopen, numareaportals * sizeof(portalopen[0]), 1, f);
}

/*
===================
CM_ReadPortalState

Reads the portal state from a savegame file
and recalculates the area connections
===================
*/
void CM_ReadPortalState (FILE *f)
{
	int num;
	
	fread(&num,1,sizeof(num),f);
	if ( num != numareaportals ) {
		Com_Error (ERR_DROP, "numareaportals differs from save game");
	}
	fread(portalopen,1, num * sizeof(portalopen[0]),f);
	FloodAreaConnections ();
}

/*
=============
CM_HeadnodeVisible

Returns true if any leaf under headnode has a cluster that
is potentially visible
=============
*/
qboolean CM_HeadnodeVisible (int nodenum, byte *visbits)
{
	int		leafnum;
	int		cluster;
	cnode_t	*node;

	if (nodenum < 0)
	{
		leafnum = -1-nodenum;
		cluster = map_leafs[leafnum].cluster;
		if (cluster == -1)
			return false;
		if (visbits[cluster>>3] & (1<<(cluster&7)))
			return true;
		return false;
	}

	node = &map_nodes[nodenum];
	if (CM_HeadnodeVisible(node->children[0], visbits))
		return true;
	return CM_HeadnodeVisible(node->children[1], visbits);
}

void CM_Release(void)
{
	CMod_Release();
}
