//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/r_poly.c                                               $
// $Revision:: 3                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 8/8/01 4:42 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// 

#include <assert.h>
#include "r_local.h"

#define AFFINE_SPANLET_SIZE      16
#define AFFINE_SPANLET_SIZE_BITS 4

typedef struct
{
	byte     *pbase, *pdest;
	unsigned short *pbase16, *pdest16;
	int alphalevel;
	short	 *pz;
	fixed16_t s, t;
	fixed16_t sstep, tstep;
	int       izi, izistep, izistep_times_2;
	int       spancount;
	unsigned  u, v;
} spanletvars_t;

spanletvars_t s_spanletvars;

static int r_polyblendcolor;
static int r_polyblendcolor3[3];
static byte * r_polycolormap = NULL;
static unsigned short * r_polycolormap16 = NULL;
static qboolean r_polynonlit;

static espan_t	*s_polygon_spans;

polydesc_t	r_polydesc;

msurface_t *r_alpha_surfaces;

extern int *r_turb_turb;

static int		clip_current;
vec5_t	r_clip_verts[2][MAXWORKINGVERTS+2];

static int		s_minindex, s_maxindex;

static void R_DrawPoly( qboolean iswater );

/*
** R_DrawSpanletOpaque
*/
static void R_DrawSpanletOpaque( void )
{
	unsigned btemp;

	do
	{
		unsigned ts, tt;

		ts = s_spanletvars.s >> 16;
		tt = s_spanletvars.t >> 16;

		btemp = *(s_spanletvars.pbase + (ts) + (tt) * cachewidth);
		if (btemp != 255)
		{
			if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
			{
				*s_spanletvars.pz    = s_spanletvars.izi >> 16;
				*s_spanletvars.pdest = btemp;
			}
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;
	} while (--s_spanletvars.spancount > 0);
}

/*
** R_DrawSpanletNonlit
*/
static void R_DrawSpanletNonlit( void )
{
	unsigned btemp;

	do
	{
		unsigned ts, tt;

		ts = s_spanletvars.s >> 16;
		ts &= r_polydesc.pixel_width-1;
		tt = s_spanletvars.t >> 16;
		tt &= r_polydesc.pixel_height-1;

		btemp = *(s_spanletvars.pbase + (ts) + (tt) * cachewidth);
		if (btemp != 255)
		{
			if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
			{
				*s_spanletvars.pz    = s_spanletvars.izi >> 16;
				*s_spanletvars.pdest = r_polycolormap[ btemp ];
			}
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;
	} while (--s_spanletvars.spancount > 0);
}

/*
** R_DrawSpanletNonlit_Alpha
*/
static void R_DrawSpanletNonlit_Alpha( void )
{
	unsigned btemp;

	if ( s_spanletvars.alphalevel > 33 )
	{
		do
		{
			unsigned ts, tt;

			ts = s_spanletvars.s >> 16;
			ts &= r_polydesc.pixel_width-1;
			tt = s_spanletvars.t >> 16;
			tt &= r_polydesc.pixel_height-1;

			btemp = *(s_spanletvars.pbase + (ts) + (tt) * cachewidth);

			if ( btemp != 255 )
			{
				if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
				{
					*s_spanletvars.pdest = vid.alphamap[(*s_spanletvars.pdest)*256 + btemp];
				}
			}

			s_spanletvars.izi += s_spanletvars.izistep;
			s_spanletvars.pdest++;
			s_spanletvars.pz++;
			s_spanletvars.s += s_spanletvars.sstep;
			s_spanletvars.t += s_spanletvars.tstep;
		} while (--s_spanletvars.spancount > 0);
	}
	else
	{
		do
		{
			unsigned ts, tt;

			ts = s_spanletvars.s >> 16;
			ts &= r_polydesc.pixel_width-1;
			tt = s_spanletvars.t >> 16;
			tt &= r_polydesc.pixel_height-1;

			btemp = *(s_spanletvars.pbase + (ts) + (tt) * cachewidth);

			if ( btemp != 255 )
			{
				if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
				{
					*s_spanletvars.pdest = vid.alphamap[btemp*256+*s_spanletvars.pdest];
				}
			}

			s_spanletvars.izi += s_spanletvars.izistep;
			s_spanletvars.pdest++;
			s_spanletvars.pz++;
			s_spanletvars.s += s_spanletvars.sstep;
			s_spanletvars.t += s_spanletvars.tstep;
		} while (--s_spanletvars.spancount > 0);
	}
}

/*
** R_DrawSpanletTurbulentStipple33
*/
static void R_DrawSpanletTurbulentStipple33( void )
{
	unsigned btemp;
	int	     sturb, tturb;
	byte    *pdest = s_spanletvars.pdest;
	short   *pz    = s_spanletvars.pz;
	int      izi   = s_spanletvars.izi;
	
	if ( s_spanletvars.v & 1 )
	{
		s_spanletvars.pdest += s_spanletvars.spancount;
		s_spanletvars.pz    += s_spanletvars.spancount;

		if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
			s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
		else
			s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;

		if ( s_spanletvars.u & 1 )
		{
			izi += s_spanletvars.izistep;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest++;
			pz++;
			s_spanletvars.spancount--;
		}

		s_spanletvars.sstep   *= 2;
		s_spanletvars.tstep   *= 2;

		while ( s_spanletvars.spancount > 0 )
		{
			sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
			tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;

			btemp = *( s_spanletvars.pbase + ( sturb ) + ( tturb << 6 ) );

			if ( *pz <= ( izi >> 16 ) )
				*pdest = btemp;

			izi               += s_spanletvars.izistep_times_2;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest += 2;
			pz    += 2;

			s_spanletvars.spancount -= 2;
		}
	}
}

/*
** R_DrawSpanletTurbulentStipple66
*/
static void R_DrawSpanletTurbulentStipple66( void )
{
	unsigned btemp;
	int	     sturb, tturb;
	byte    *pdest = s_spanletvars.pdest;
	short   *pz    = s_spanletvars.pz;
	int      izi   = s_spanletvars.izi;
	
	if ( !( s_spanletvars.v & 1 ) )
	{
		s_spanletvars.pdest += s_spanletvars.spancount;
		s_spanletvars.pz    += s_spanletvars.spancount;

		if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
			s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
		else
			s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;
		
		if ( s_spanletvars.u & 1 )
		{
			izi += s_spanletvars.izistep;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest++;
			pz++;
			s_spanletvars.spancount--;
		}

		s_spanletvars.sstep   *= 2;
		s_spanletvars.tstep   *= 2;

		while ( s_spanletvars.spancount > 0 )
		{
			sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
			tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;
			
			btemp = *( s_spanletvars.pbase + ( sturb ) + ( tturb << 6 ) );
			
			if ( *pz <= ( izi >> 16 ) )
				*pdest = btemp;
			
			izi               += s_spanletvars.izistep_times_2;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;
			
			pdest += 2;
			pz    += 2;
			
			s_spanletvars.spancount -= 2;
		}
	}
	else
	{
		s_spanletvars.pdest += s_spanletvars.spancount;
		s_spanletvars.pz    += s_spanletvars.spancount;

		if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
			s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
		else
			s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;
		
		while ( s_spanletvars.spancount > 0 )
		{
			sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
			tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;
			
			btemp = *( s_spanletvars.pbase + ( sturb ) + ( tturb << 6 ) );
			
			if ( *pz <= ( izi >> 16 ) )
				*pdest = btemp;
			
			izi               += s_spanletvars.izistep;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;
			
			pdest++;
			pz++;
			
			s_spanletvars.spancount--;
		}
	}
}

/*
** R_DrawSpanletTurbulentBlended
*/
static void R_DrawSpanletTurbulentBlended66( void )
{
	unsigned btemp;
	int	     sturb, tturb;

	do
	{
		sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
		tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;

		btemp = *( s_spanletvars.pbase + ( sturb ) + ( tturb << 6 ) );

		if ( *s_spanletvars.pz <= ( s_spanletvars.izi >> 16 ) )
			*s_spanletvars.pdest = vid.alphamap[btemp*256+*s_spanletvars.pdest];

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;

	} while ( --s_spanletvars.spancount > 0 );
}

static void R_DrawSpanletTurbulentBlended33( void )
{
	unsigned btemp;
	int	     sturb, tturb;

	do
	{
		sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
		tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;

		btemp = *( s_spanletvars.pbase + ( sturb ) + ( tturb << 6 ) );

		if ( *s_spanletvars.pz <= ( s_spanletvars.izi >> 16 ) )
			*s_spanletvars.pdest = vid.alphamap[btemp+*s_spanletvars.pdest*256];

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;

	} while ( --s_spanletvars.spancount > 0 );
}

/*
** R_DrawSpanlet33
*/
static void R_DrawSpanlet33( void )
{
	unsigned btemp;

	do
	{
		unsigned ts, tt;

		ts = s_spanletvars.s >> 16;
		tt = s_spanletvars.t >> 16;

		btemp = *(s_spanletvars.pbase + (ts) + (tt) * cachewidth);

		if ( btemp != 255 )
		{
			if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
			{
				*s_spanletvars.pdest = vid.alphamap[btemp+*s_spanletvars.pdest*256];
			}
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;
	} while (--s_spanletvars.spancount > 0);
}

static void R_DrawSpanletConstant33( void )
{
	do
	{
		if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
		{
			*s_spanletvars.pdest = vid.alphamap[r_polyblendcolor+*s_spanletvars.pdest*256];
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest++;
		s_spanletvars.pz++;
	} while (--s_spanletvars.spancount > 0);
}

static void R_DrawSpanletConstant66( void )
{
	byte * alphamap;

	alphamap = vid.alphamap + ( r_polyblendcolor * 256 );
	do
	{
		if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
		{
			*s_spanletvars.pdest = alphamap[ *s_spanletvars.pdest ];
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest++;
		s_spanletvars.pz++;
	} while (--s_spanletvars.spancount > 0);
}

/*
** R_DrawSpanlet66
*/
static void R_DrawSpanlet66( void )
{
	unsigned btemp;

	do
	{
		unsigned ts, tt;

		ts = s_spanletvars.s >> 16;
		tt = s_spanletvars.t >> 16;

		btemp = *(s_spanletvars.pbase + (ts) + (tt) * cachewidth);

		if ( btemp != 255 )
		{
			if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
			{
				*s_spanletvars.pdest = vid.alphamap[btemp*256+*s_spanletvars.pdest];
			}
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;
	} while (--s_spanletvars.spancount > 0);
}

/*
** R_DrawSpanlet33Stipple
*/
static void R_DrawSpanlet33Stipple( void )
{
	unsigned btemp;
	byte    *pdest = s_spanletvars.pdest;
	short   *pz    = s_spanletvars.pz;
	int      izi   = s_spanletvars.izi;
	
	if ( r_polydesc.stipple_parity ^ ( s_spanletvars.v & 1 ) )
	{
		s_spanletvars.pdest += s_spanletvars.spancount;
		s_spanletvars.pz    += s_spanletvars.spancount;

		if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
			s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
		else
			s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;
		
		if ( r_polydesc.stipple_parity ^ ( s_spanletvars.u & 1 ) )
		{
			izi += s_spanletvars.izistep;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest++;
			pz++;
			s_spanletvars.spancount--;
		}

		s_spanletvars.sstep *= 2;
		s_spanletvars.tstep *= 2;

		while ( s_spanletvars.spancount > 0 )
		{
			unsigned s = s_spanletvars.s >> 16;
			unsigned t = s_spanletvars.t >> 16;

			btemp = *( s_spanletvars.pbase + ( s ) + ( t * cachewidth ) );
			
			if ( btemp != 255 )
			{
				if ( *pz <= ( izi >> 16 ) )
					*pdest = btemp;
			}
			
			izi               += s_spanletvars.izistep_times_2;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;
			
			pdest += 2;
			pz    += 2;
			
			s_spanletvars.spancount -= 2;
		}
	}
}

/*
** R_DrawSpanlet66Stipple
*/
static void R_DrawSpanlet66Stipple( void )
{
	unsigned btemp;
	byte    *pdest = s_spanletvars.pdest;
	short   *pz    = s_spanletvars.pz;
	int      izi   = s_spanletvars.izi;

	s_spanletvars.pdest += s_spanletvars.spancount;
	s_spanletvars.pz    += s_spanletvars.spancount;

	if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
		s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
	else
		s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;

	if ( r_polydesc.stipple_parity ^ ( s_spanletvars.v & 1 ) )
	{
		if ( r_polydesc.stipple_parity ^ ( s_spanletvars.u & 1 ) )
		{
			izi += s_spanletvars.izistep;
			s_spanletvars.s += s_spanletvars.sstep;
			s_spanletvars.t += s_spanletvars.tstep;

			pdest++;
			pz++;
			s_spanletvars.spancount--;
		}

		s_spanletvars.sstep *= 2;
		s_spanletvars.tstep *= 2;

		while ( s_spanletvars.spancount > 0 )
		{
			unsigned s = s_spanletvars.s >> 16;
			unsigned t = s_spanletvars.t >> 16;
			
			btemp = *( s_spanletvars.pbase + ( s ) + ( t * cachewidth ) );

			if ( btemp != 255 )
			{
				if ( *pz <= ( izi >> 16 ) )
					*pdest = btemp;
			}
			
			izi             += s_spanletvars.izistep_times_2;
			s_spanletvars.s += s_spanletvars.sstep;
			s_spanletvars.t += s_spanletvars.tstep;
			
			pdest += 2;
			pz    += 2;
			
			s_spanletvars.spancount -= 2;
		}
	}
	else
	{
		while ( s_spanletvars.spancount > 0 )
		{
			unsigned s = s_spanletvars.s >> 16;
			unsigned t = s_spanletvars.t >> 16;
			
			btemp = *( s_spanletvars.pbase + ( s ) + ( t * cachewidth ) );
			
			if ( btemp != 255 )
			{
				if ( *pz <= ( izi >> 16 ) )
					*pdest = btemp;
			}
			
			izi             += s_spanletvars.izistep;
			s_spanletvars.s += s_spanletvars.sstep;
			s_spanletvars.t += s_spanletvars.tstep;
			
			pdest++;
			pz++;
			
			s_spanletvars.spancount--;
		}
	}
}

/*
** R_DrawSpanletOpaque_16bit
*/
static void R_DrawSpanletOpaque_16bit( void )
{
	unsigned btemp;

	do
	{
		unsigned ts, tt;

		ts = s_spanletvars.s >> 16;
		tt = s_spanletvars.t >> 16;

		btemp = *(s_spanletvars.pbase16 + (ts) + (tt) * cachewidth);
		if (btemp != TRANSPARENT_COLOR16)
		{
			if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
			{
				*s_spanletvars.pz    = s_spanletvars.izi >> 16;
				*s_spanletvars.pdest16 = btemp;
			}
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest16++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;
	} while (--s_spanletvars.spancount > 0);
}


/*
** R_DrawSpanletNonlit_16bit
*/
static void R_DrawSpanletNonlit_16bit( void )
{
	byte *psrc;
	unsigned btemp;

	psrc = (byte *)s_spanletvars.pbase16;
	do
	{
		unsigned ts, tt;

		ts = s_spanletvars.s >> 16;
		ts &= r_polydesc.pixel_width-1;
		tt = s_spanletvars.t >> 16;
		tt &= r_polydesc.pixel_height-1;

		btemp = *(psrc + (ts) + (tt) * cachewidth);
		if (btemp != 255)
		{
			if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
			{
				*s_spanletvars.pz    = s_spanletvars.izi >> 16;
				*s_spanletvars.pdest16 = r_polycolormap16[ btemp ];
			}
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest16++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;
	} while (--s_spanletvars.spancount > 0);
}

/*
** R_DrawSpanletNonlit_16bit_Alpha
*/
static void R_DrawSpanletNonlit_16bit_Alpha( void )
{
	byte * psrc;
	unsigned btemp;

	psrc = (byte *)s_spanletvars.pbase16;

	do
	{
		unsigned ts, tt;

		ts = s_spanletvars.s >> 16;
		ts &= r_polydesc.pixel_width-1;
		tt = s_spanletvars.t >> 16;
		tt &= r_polydesc.pixel_height-1;

		btemp = *(psrc + (ts) + (tt) * cachewidth);

		if ( btemp != TRANSPARENT_COLOR16 )
		{
			if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
			{
				int sr,sg,sb;
				int dr,dg,db;
				int dst;

				sr = UNPACKRED16BIT( btemp );
				sg = UNPACKGREEN16BIT( btemp );
				sb = UNPACKBLUE16BIT( btemp );
				dst = *s_spanletvars.pdest16;
				dr = UNPACKRED16BIT( dst );
				dg = UNPACKGREEN16BIT( dst );
				db = UNPACKBLUE16BIT( dst );
				dr = INTREDALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sr, dr );
				dg = INTGREENALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sg, dg );
				db = INTBLUEALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sb, db );
				*s_spanletvars.pdest16 = PACKALPHA16BIT( dr,dg,db );
			}
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest16++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;
	} while (--s_spanletvars.spancount > 0);
}

/*
** R_DrawSpanletTurbulentStipple33_16bit
*/
static void R_DrawSpanletTurbulentStipple33_16bit( void )
{
	unsigned btemp;
	int sturb, tturb;
	unsigned short *pdest = s_spanletvars.pdest16;
	short   *pz    = s_spanletvars.pz;
	int izi   = s_spanletvars.izi;

	if ( s_spanletvars.v & 1 )
	{
		s_spanletvars.pdest16 += s_spanletvars.spancount;
		s_spanletvars.pz    += s_spanletvars.spancount;

		if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
			s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
		else
			s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;

		if ( s_spanletvars.u & 1 )
		{
			izi += s_spanletvars.izistep;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest++;
			pz++;
			s_spanletvars.spancount--;
		}

		s_spanletvars.sstep   *= 2;
		s_spanletvars.tstep   *= 2;

		while ( s_spanletvars.spancount > 0 )
		{
			sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
			tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;

			btemp = *( s_spanletvars.pbase16 + ( sturb ) + ( tturb << 6 ) );

			if ( *pz <= ( izi >> 16 ) )
				*pdest = btemp;

			izi               += s_spanletvars.izistep_times_2;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest += 2;
			pz    += 2;

			s_spanletvars.spancount -= 2;
		}
	}
}

/*
** R_DrawSpanletTurbulentStipple66_16bit
*/
static void R_DrawSpanletTurbulentStipple66_16bit( void )
{
	unsigned btemp;
	int sturb, tturb;
	unsigned short *pdest = s_spanletvars.pdest16;
	short   *pz    = s_spanletvars.pz;
	int izi   = s_spanletvars.izi;

	if ( !( s_spanletvars.v & 1 ) )
	{
		s_spanletvars.pdest16 += s_spanletvars.spancount;
		s_spanletvars.pz    += s_spanletvars.spancount;

		if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
			s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
		else
			s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;

		if ( s_spanletvars.u & 1 )
		{
			izi += s_spanletvars.izistep;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest++;
			pz++;
			s_spanletvars.spancount--;
		}

		s_spanletvars.sstep   *= 2;
		s_spanletvars.tstep   *= 2;

		while ( s_spanletvars.spancount > 0 )
		{
			sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
			tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;

			btemp = *( s_spanletvars.pbase16 + ( sturb ) + ( tturb << 6 ) );

			if ( *pz <= ( izi >> 16 ) )
				*pdest = btemp;

			izi               += s_spanletvars.izistep_times_2;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest += 2;
			pz    += 2;

			s_spanletvars.spancount -= 2;
		}
	}
	else
	{
		s_spanletvars.pdest16 += s_spanletvars.spancount;
		s_spanletvars.pz    += s_spanletvars.spancount;

		if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
			s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
		else
			s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;

		while ( s_spanletvars.spancount > 0 )
		{
			sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
			tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;

			btemp = *( s_spanletvars.pbase16 + ( sturb ) + ( tturb << 6 ) );

			if ( *pz <= ( izi >> 16 ) )
				*pdest = btemp;

			izi               += s_spanletvars.izistep;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest++;
			pz++;

			s_spanletvars.spancount--;
		}
	}
}

/*
** R_DrawSpanletTurbulentBlendedAlpha_16bit
*/
static void R_DrawSpanletTurbulentBlendedAlpha_16bit( void )
{
	unsigned btemp;
	int sturb, tturb;

	do
	{
		sturb = ((s_spanletvars.s + r_turb_turb[(s_spanletvars.t>>16)&(CYCLE-1)])>>16)&63;
		tturb = ((s_spanletvars.t + r_turb_turb[(s_spanletvars.s>>16)&(CYCLE-1)])>>16)&63;

		btemp = *( s_spanletvars.pbase16 + ( sturb ) + ( tturb << 6 ) );

		if ( *s_spanletvars.pz <= ( s_spanletvars.izi >> 16 ) )
		{
			int sr,sg,sb;
			int dr,dg,db;
			int dst;

			sr = UNPACKRED16BIT( btemp );
			sg = UNPACKGREEN16BIT( btemp );
			sb = UNPACKBLUE16BIT( btemp );
			dst = *s_spanletvars.pdest16;
			dr = UNPACKRED16BIT( dst );
			dg = UNPACKGREEN16BIT( dst );
			db = UNPACKBLUE16BIT( dst );
			dr = INTREDALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sr, dr );
			dg = INTGREENALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sg, dg );
			db = INTBLUEALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sb, db );
			*s_spanletvars.pdest16 = PACKALPHA16BIT( dr,dg,db );
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest16++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;

	} while ( --s_spanletvars.spancount > 0 );
}

/*
** R_DrawSpanletAlpha_16bit
*/
static void R_DrawSpanletAlpha_16bit( void )
{
	unsigned btemp;

	do
	{
		unsigned ts, tt;

		ts = s_spanletvars.s >> 16;
		tt = s_spanletvars.t >> 16;

		btemp = *(s_spanletvars.pbase16 + (ts) + (tt) * cachewidth);

		if ( btemp != TRANSPARENT_COLOR16 )
		{
			if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
			{
				int sr,sg,sb;
				int dr,dg,db;
				int dst;

				sr = UNPACKRED16BIT( btemp );
				sg = UNPACKGREEN16BIT( btemp );
				sb = UNPACKBLUE16BIT( btemp );
				dst = *s_spanletvars.pdest16;
				dr = UNPACKRED16BIT( dst );
				dg = UNPACKGREEN16BIT( dst );
				db = UNPACKBLUE16BIT( dst );
				dr = INTREDALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sr, dr );
				dg = INTGREENALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sg, dg );
				db = INTBLUEALPHA16( vid.alphamap16, s_spanletvars.alphalevel, sb, db );
				*s_spanletvars.pdest16 = PACKALPHA16BIT( dr,dg,db );
			}
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest16++;
		s_spanletvars.pz++;
		s_spanletvars.s += s_spanletvars.sstep;
		s_spanletvars.t += s_spanletvars.tstep;
	} while (--s_spanletvars.spancount > 0);
}

static void R_DrawSpanletConstantAlpha_16bit( void )
{
	do
	{
		if (*s_spanletvars.pz <= (s_spanletvars.izi >> 16))
		{
			int dr,dg,db;
			int dst;

			dst = *s_spanletvars.pdest16;
			dr = UNPACKRED16BIT( dst );
			dg = UNPACKGREEN16BIT( dst );
			db = UNPACKBLUE16BIT( dst );
			dr = INTREDALPHA16( vid.alphamap16, s_spanletvars.alphalevel, r_polyblendcolor3[0], dr );
			dg = INTGREENALPHA16( vid.alphamap16, s_spanletvars.alphalevel, r_polyblendcolor3[1], dg );
			db = INTBLUEALPHA16( vid.alphamap16, s_spanletvars.alphalevel, r_polyblendcolor3[2], db );
			*s_spanletvars.pdest16 = PACKALPHA16BIT( dr,dg,db );
		}

		s_spanletvars.izi += s_spanletvars.izistep;
		s_spanletvars.pdest16++;
		s_spanletvars.pz++;
	} while (--s_spanletvars.spancount > 0);
}

/*
** R_DrawSpanlet33Stipple
*/
static void R_DrawSpanlet33Stipple_16bit( void )
{
	unsigned btemp;
	unsigned short *pdest = s_spanletvars.pdest16;
	short   *pz    = s_spanletvars.pz;
	int izi   = s_spanletvars.izi;

	if ( r_polydesc.stipple_parity ^ ( s_spanletvars.v & 1 ) )
	{
		s_spanletvars.pdest16 += s_spanletvars.spancount;
		s_spanletvars.pz    += s_spanletvars.spancount;

		if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
			s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
		else
			s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;

		if ( r_polydesc.stipple_parity ^ ( s_spanletvars.u & 1 ) )
		{
			izi += s_spanletvars.izistep;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest++;
			pz++;
			s_spanletvars.spancount--;
		}

		s_spanletvars.sstep *= 2;
		s_spanletvars.tstep *= 2;

		while ( s_spanletvars.spancount > 0 )
		{
			unsigned s = s_spanletvars.s >> 16;
			unsigned t = s_spanletvars.t >> 16;

			btemp = *( s_spanletvars.pbase16 + ( s ) + ( t * cachewidth ) );

			if ( btemp != TRANSPARENT_COLOR16 )
			{
				if ( *pz <= ( izi >> 16 ) )
					*pdest = btemp;
			}

			izi               += s_spanletvars.izistep_times_2;
			s_spanletvars.s   += s_spanletvars.sstep;
			s_spanletvars.t   += s_spanletvars.tstep;

			pdest += 2;
			pz    += 2;

			s_spanletvars.spancount -= 2;
		}
	}
}

/*
** R_DrawSpanlet66Stipple_16bit
*/
static void R_DrawSpanlet66Stipple_16bit( void )
{
	unsigned btemp;
	unsigned short *pdest = s_spanletvars.pdest16;
	short   *pz    = s_spanletvars.pz;
	int izi   = s_spanletvars.izi;

	s_spanletvars.pdest16 += s_spanletvars.spancount;
	s_spanletvars.pz    += s_spanletvars.spancount;

	if ( s_spanletvars.spancount == AFFINE_SPANLET_SIZE )
		s_spanletvars.izi += s_spanletvars.izistep << AFFINE_SPANLET_SIZE_BITS;
	else
		s_spanletvars.izi += s_spanletvars.izistep * s_spanletvars.izistep;

	if ( r_polydesc.stipple_parity ^ ( s_spanletvars.v & 1 ) )
	{
		if ( r_polydesc.stipple_parity ^ ( s_spanletvars.u & 1 ) )
		{
			izi += s_spanletvars.izistep;
			s_spanletvars.s += s_spanletvars.sstep;
			s_spanletvars.t += s_spanletvars.tstep;

			pdest++;
			pz++;
			s_spanletvars.spancount--;
		}

		s_spanletvars.sstep *= 2;
		s_spanletvars.tstep *= 2;

		while ( s_spanletvars.spancount > 0 )
		{
			unsigned s = s_spanletvars.s >> 16;
			unsigned t = s_spanletvars.t >> 16;

			btemp = *( s_spanletvars.pbase16 + ( s ) + ( t * cachewidth ) );

			if ( btemp != TRANSPARENT_COLOR16 )
			{
				if ( *pz <= ( izi >> 16 ) )
					*pdest = btemp;
			}

			izi             += s_spanletvars.izistep_times_2;
			s_spanletvars.s += s_spanletvars.sstep;
			s_spanletvars.t += s_spanletvars.tstep;

			pdest += 2;
			pz    += 2;

			s_spanletvars.spancount -= 2;
		}
	}
	else
	{
		while ( s_spanletvars.spancount > 0 )
		{
			unsigned s = s_spanletvars.s >> 16;
			unsigned t = s_spanletvars.t >> 16;

			btemp = *( s_spanletvars.pbase16 + ( s ) + ( t * cachewidth ) );

			if ( btemp != TRANSPARENT_COLOR16 )
			{
				if ( *pz <= ( izi >> 16 ) )
					*pdest = btemp;
			}

			izi             += s_spanletvars.izistep;
			s_spanletvars.s += s_spanletvars.sstep;
			s_spanletvars.t += s_spanletvars.tstep;

			pdest++;
			pz++;

			s_spanletvars.spancount--;
		}
	}
}

/*
** R_ClipPolyFace
**
** Clips the winding at clip_verts[clip_current] and changes clip_current
** Throws out the back side
*/
static int R_ClipPolyFace (int nump, clipplane_t *pclipplane)
{
	int		i, outcount;
	float	dists[MAXWORKINGVERTS+3];
	float	frac, clipdist, *pclipnormal;
	float	*in, *instep, *outstep, *vert2;

	clipdist = pclipplane->dist;
	pclipnormal = pclipplane->normal;
	
// calc dists
	if (clip_current)
	{
		in = r_clip_verts[1][0];
		outstep = r_clip_verts[0][0];
		clip_current = 0;
	}
	else
	{
		in = r_clip_verts[0][0];
		outstep = r_clip_verts[1][0];
		clip_current = 1;
	}
	
	instep = in;
	for (i=0 ; i<nump ; i++, instep += sizeof (vec5_t) / sizeof (float))
	{
		dists[i] = DotProduct (instep, pclipnormal) - clipdist;
	}
	
// handle wraparound case
	dists[nump] = dists[0];
	memcpy(instep, in, sizeof (vec5_t));


// clip the winding
	instep = in;
	outcount = 0;

	for (i=0 ; i<nump ; i++, instep += sizeof (vec5_t) / sizeof (float))
	{
		if (dists[i] >= 0)
		{
			memcpy(outstep, instep, sizeof (vec5_t));
			outstep += sizeof (vec5_t) / sizeof (float);
			outcount++;
		}

		if (dists[i] == 0 || dists[i+1] == 0)
			continue;

		if ( (dists[i] > 0) == (dists[i+1] > 0) )
			continue;
			
	// split it into a new vertex
		frac = dists[i] / (dists[i] - dists[i+1]);
			
		vert2 = instep + sizeof (vec5_t) / sizeof (float);
		
		outstep[0] = instep[0] + frac*(vert2[0] - instep[0]);
		outstep[1] = instep[1] + frac*(vert2[1] - instep[1]);
		outstep[2] = instep[2] + frac*(vert2[2] - instep[2]);
		outstep[3] = instep[3] + frac*(vert2[3] - instep[3]);
		outstep[4] = instep[4] + frac*(vert2[4] - instep[4]);

		outstep += sizeof (vec5_t) / sizeof (float);
		outcount++;
	}	
	
	return outcount;
}

/*
** R_PolygonDrawSpans
*/
static void R_PolygonDrawSpans(espan_t *pspan, qboolean iswater )
{
	int			count;
	fixed16_t	snext, tnext;
	float		sdivz, tdivz, zi, z, du, dv, spancountminus1;
	float		sdivzspanletstepu, tdivzspanletstepu, zispanletstepu;

	if ( vid.bitdepth > 8 )
	{
		s_spanletvars.pbase = NULL;
		s_spanletvars.pbase16 = ( unsigned short * )cacheblock;
	}
	else
	{
		s_spanletvars.pbase = cacheblock;
	}

	if ( iswater )
		r_turb_turb = sintable + ((int)(r_newrefdef.time*SPEED)&(CYCLE-1));

	sdivzspanletstepu = d_sdivzstepu * AFFINE_SPANLET_SIZE;
	tdivzspanletstepu = d_tdivzstepu * AFFINE_SPANLET_SIZE;
	zispanletstepu = d_zistepu * AFFINE_SPANLET_SIZE;

// we count on FP exceptions being turned off to avoid range problems
	s_spanletvars.izistep = (int)(d_zistepu * 0x8000 * 0x10000);
	s_spanletvars.izistep_times_2 = s_spanletvars.izistep * 2;

	s_spanletvars.pz = 0;

	do
	{
		count = pspan->count;

		if (count <= 0)
			goto NextSpan;

		if ( vid.bitdepth > 8 )
		{
			s_spanletvars.pdest   = NULL;
			s_spanletvars.pdest16   = (unsigned short *)d_viewbuffer + (d_scantable[pspan->v]>>1) + pspan->u;
		}
		else
		{
			s_spanletvars.pdest   = (byte *)d_viewbuffer + ( d_scantable[pspan->v] /*r_screenwidth * pspan->v*/) + pspan->u;
		}
		s_spanletvars.pz      = d_pzbuffer + (d_zwidth * pspan->v) + pspan->u;
		s_spanletvars.u       = pspan->u;
		s_spanletvars.v       = pspan->v;

	// calculate the initial s/z, t/z, 1/z, s, and t and clamp
		du = (float)pspan->u;
		dv = (float)pspan->v;

		sdivz = d_sdivzorigin + dv*d_sdivzstepv + du*d_sdivzstepu;
		tdivz = d_tdivzorigin + dv*d_tdivzstepv + du*d_tdivzstepu;

		zi = d_ziorigin + dv*d_zistepv + du*d_zistepu;
		z = (float)0x10000 / zi;	// prescale to 16.16 fixed-point
	// we count on FP exceptions being turned off to avoid range problems
		s_spanletvars.izi = (int)(zi * 0x8000 * 0x10000);

		s_spanletvars.s = (int)(sdivz * z) + sadjust;
		s_spanletvars.t = (int)(tdivz * z) + tadjust;

		if ( !iswater && !r_polynonlit )
		{
			if (s_spanletvars.s > bbextents)
				s_spanletvars.s = bbextents;
			else if (s_spanletvars.s < 0)
				s_spanletvars.s = 0;

			if (s_spanletvars.t > bbextentt)
				s_spanletvars.t = bbextentt;
			else if (s_spanletvars.t < 0)
				s_spanletvars.t = 0;
		}

		do
		{
		// calculate s and t at the far end of the span
			if (count >= AFFINE_SPANLET_SIZE )
				s_spanletvars.spancount = AFFINE_SPANLET_SIZE;
			else
				s_spanletvars.spancount = count;

			count -= s_spanletvars.spancount;

			if (count)
			{
			// calculate s/z, t/z, zi->fixed s and t at far end of span,
			// calculate s and t steps across span by shifting
				sdivz += sdivzspanletstepu;
				tdivz += tdivzspanletstepu;
				zi += zispanletstepu;
				z = (float)0x10000 / zi;	// prescale to 16.16 fixed-point

				snext = (int)(sdivz * z) + sadjust;
				tnext = (int)(tdivz * z) + tadjust;

				if ( !iswater && !r_polynonlit )
				{
					if (snext > bbextents)
						snext = bbextents;
					else if (snext < AFFINE_SPANLET_SIZE)
						snext = AFFINE_SPANLET_SIZE;	// prevent round-off error on <0 steps from
									//  from causing overstepping & running off the
									//  edge of the texture

					if (tnext > bbextentt)
						tnext = bbextentt;
					else if (tnext < AFFINE_SPANLET_SIZE)
						tnext = AFFINE_SPANLET_SIZE;	// guard against round-off error on <0 steps
				}

				s_spanletvars.sstep = (snext - s_spanletvars.s) >> AFFINE_SPANLET_SIZE_BITS;
				s_spanletvars.tstep = (tnext - s_spanletvars.t) >> AFFINE_SPANLET_SIZE_BITS;
			}
			else
			{
			// calculate s/z, t/z, zi->fixed s and t at last pixel in span (so
			// can't step off polygon), clamp, calculate s and t steps across
			// span by division, biasing steps low so we don't run off the
			// texture
				spancountminus1 = (float)(s_spanletvars.spancount - 1);
				sdivz += d_sdivzstepu * spancountminus1;
				tdivz += d_tdivzstepu * spancountminus1;
				zi += d_zistepu * spancountminus1;
				z = (float)0x10000 / zi;	// prescale to 16.16 fixed-point
				snext = (int)(sdivz * z) + sadjust;
				tnext = (int)(tdivz * z) + tadjust;

				if ( !iswater && !r_polynonlit )
				{
					if (snext > bbextents)
						snext = bbextents;
					else if (snext < AFFINE_SPANLET_SIZE)
						snext = AFFINE_SPANLET_SIZE;	// prevent round-off error on <0 steps from
									//  from causing overstepping & running off the
									//  edge of the texture

					if (tnext > bbextentt)
						tnext = bbextentt;
					else if (tnext < AFFINE_SPANLET_SIZE)
						tnext = AFFINE_SPANLET_SIZE;	// guard against round-off error on <0 steps
				}

				if (s_spanletvars.spancount > 1)
				{
					s_spanletvars.sstep = (snext - s_spanletvars.s) / (s_spanletvars.spancount - 1);
					s_spanletvars.tstep = (tnext - s_spanletvars.t) / (s_spanletvars.spancount - 1);
				}
			}

			if ( iswater )
			{
				s_spanletvars.s = s_spanletvars.s & ((CYCLE<<16)-1);
				s_spanletvars.t = s_spanletvars.t & ((CYCLE<<16)-1);
			}

			r_polydesc.drawspanlet();

			s_spanletvars.s = snext;
			s_spanletvars.t = tnext;

		} while (count > 0);

NextSpan:
		pspan++;

	} while (pspan->count != DS_SPAN_LIST_END);
}

/*
**
** R_PolygonScanLeftEdge
**
** Goes through the polygon and scans the left edge, filling in 
** screen coordinate data for the spans
*/
static void R_PolygonScanLeftEdge (void)
{
	int			i, v, itop, ibottom, lmaxindex;
	emitpoint_t	*pvert, *pnext;
	espan_t		*pspan;
	float		du, dv, vtop, vbottom, slope;
	fixed16_t	u, u_step;

	pspan = s_polygon_spans;
	i = s_minindex;
	if (i == 0)
		i = r_polydesc.nump;

	lmaxindex = s_maxindex;
	if (lmaxindex == 0)
		lmaxindex = r_polydesc.nump;

	vtop = ceil (r_polydesc.pverts[i].v);

	do
	{
		pvert = &r_polydesc.pverts[i];
		pnext = pvert - 1;

		vbottom = ceil (pnext->v);

      assert( vtop >= 0 );
      assert( vtop <= vid.height );
      assert( vbottom >= 0 );
      assert( vbottom <= vid.height );

		if (vtop < vbottom)
		{
			du = pnext->u - pvert->u;
			dv = pnext->v - pvert->v;

			slope = du / dv;
			u_step = (int)(slope * 0x10000);
		// adjust u to ceil the integer portion
			u = (int)((pvert->u + (slope * (vtop - pvert->v))) * 0x10000) +
					(0x10000 - 1);
			itop = (int)vtop;
			ibottom = (int)vbottom;

			for (v=itop ; v<ibottom ; v++)
			{
				pspan->u = u >> 16;
				pspan->v = v;

				/* We zero out count here so that when we call R_PolygonScanRightEdge, */
				/* we won't write past the end of our spans created by the left edge. */
				/* This also has the effect of fixing it in case the left edge is */
				/* mysteriously longer than the right edge.  Normally, neither of these */
				/* cases should be possible, but since the tools sometimes create concave */
				/* polys, this can happen and when it does, it's bad. */
				pspan->count = 0;

				u += u_step;
				pspan++;
			}
		}

		vtop = vbottom;

		i--;
		if (i == 0)
			i = r_polydesc.nump;

	} while (i != lmaxindex);

	pspan->count = DS_SPAN_LIST_END;	// mark the end of the span list 
}

/*
** R_PolygonScanRightEdge
**
** Goes through the polygon and scans the right edge, filling in
** count values.
*/
static void R_PolygonScanRightEdge (void)
{
	int			i, v, itop, ibottom;
	emitpoint_t	*pvert, *pnext;
	espan_t		*pspan;
	float		du, dv, vtop, vbottom, slope, uvert, unext, vvert, vnext;
	fixed16_t	u, u_step;

	pspan = s_polygon_spans;
	i = s_minindex;

	vvert = r_polydesc.pverts[i].v;
	if (vvert < r_refdef.fvrecty_adj)
		vvert = r_refdef.fvrecty_adj;
	if (vvert > r_refdef.fvrectbottom_adj)
		vvert = r_refdef.fvrectbottom_adj;

	vtop = ceil (vvert);

	do
	{
		pvert = &r_polydesc.pverts[i];
		pnext = pvert + 1;

		vnext = pnext->v;
		if (vnext < r_refdef.fvrecty_adj)
			vnext = r_refdef.fvrecty_adj;
		if (vnext > r_refdef.fvrectbottom_adj)
			vnext = r_refdef.fvrectbottom_adj;

		vbottom = ceil (vnext);

      assert( vtop >= 0 );
      assert( vtop <= vid.height );
      assert( vbottom >= 0 );
      assert( vbottom <= vid.height );

		if (vtop < vbottom)
		{
			uvert = pvert->u;
			if (uvert < r_refdef.fvrectx_adj)
				uvert = r_refdef.fvrectx_adj;
			if (uvert > r_refdef.fvrectright_adj)
				uvert = r_refdef.fvrectright_adj;

			unext = pnext->u;
			if (unext < r_refdef.fvrectx_adj)
				unext = r_refdef.fvrectx_adj;
			if (unext > r_refdef.fvrectright_adj)
				unext = r_refdef.fvrectright_adj;

			du = unext - uvert;
			dv = vnext - vvert;
			slope = du / dv;
			u_step = (int)(slope * 0x10000);
		// adjust u to ceil the integer portion
			u = (int)((uvert + (slope * (vtop - vvert))) * 0x10000) +
					(0x10000 - 1);
			itop = (int)vtop;
			ibottom = (int)vbottom;

			for (v=itop ; v<ibottom ; v++)
			{
				if ( pspan->count )
				{
					/* we zero out count in R_PolygonScanLeftEdge, so any value here */
					/* indicates we've passed the end of the spans created by the left */
					/* edge.  Normally, this isn't possible, but since the tools sometimes */
					/* create concave polys, this can happen and when it does, it's bad. */
					return;
				}

				pspan->count = (u >> 16) - pspan->u;
				u += u_step;
				pspan++;
			}
		}

		vtop = vbottom;
		vvert = vnext;

		i++;
		if (i == r_polydesc.nump)
			i = 0;

	} while (i != s_maxindex);

	pspan->count = DS_SPAN_LIST_END;	// mark the end of the span list 
}

/*
** R_ClipAndDrawPoly
*/
void R_ClipAndDrawPoly( float alpha, qboolean isturbulent, qboolean textured )
{
	emitpoint_t outverts[MAXWORKINGVERTS+3], *pout;
	vec_t *pv;
	int i, nump;
	float scale;
	vec3_t transformed, local;

	if ( vid.bitdepth > 8 )
	{
		if ( !textured )
		{
			s_spanletvars.alphalevel = alpha * NUMALPHALEVELS;
			if ( s_spanletvars.alphalevel >= NUMALPHALEVELS )
				s_spanletvars.alphalevel = NUMALPHALEVELS - 1;
			r_polydesc.drawspanlet = R_DrawSpanletConstantAlpha_16bit;
		}
		else
		{

			/*
			** choose the correct spanlet routine based on alpha
			*/
			if ( alpha == 1 )
			{
				//if ( fa->texinfo->flags & (SURF_WARP|SURF_WAVY|SURF_NONLIT) )
				if ( r_polynonlit )
				{
					r_polydesc.drawspanlet = R_DrawSpanletNonlit_16bit;
				}
				else
				{
					// isturbulent is ignored because we know that turbulent surfaces
					// can't be opaque
					r_polydesc.drawspanlet = R_DrawSpanletOpaque_16bit;
				}
			}
			else
			{
				s_spanletvars.alphalevel = alpha * NUMALPHALEVELS;
				if ( s_spanletvars.alphalevel >= NUMALPHALEVELS )
					s_spanletvars.alphalevel = NUMALPHALEVELS - 1;
				if ( r_polynonlit )
				{
					r_polydesc.drawspanlet = R_DrawSpanletNonlit_16bit_Alpha;
				}
				else if ( sw_stipplealpha->value )
				{
					if ( isturbulent )
					{
						if ( alpha > 0.33 )
							r_polydesc.drawspanlet = R_DrawSpanletTurbulentStipple66_16bit;
						else
							r_polydesc.drawspanlet = R_DrawSpanletTurbulentStipple33_16bit;
					}
					else
					{
						if ( alpha > 0.33 )
							r_polydesc.drawspanlet = R_DrawSpanlet66Stipple_16bit;
						else
							r_polydesc.drawspanlet = R_DrawSpanlet33Stipple_16bit;
					}
				}
				else
				{
					if ( isturbulent )
					{
						r_polydesc.drawspanlet = R_DrawSpanletTurbulentBlendedAlpha_16bit;
					}
					else
					{
						r_polydesc.drawspanlet = R_DrawSpanletAlpha_16bit;
					}
				}
			}
		}
	}
	else
	{
		if ( !textured )
		{
			if ( alpha > 0.33 )
				r_polydesc.drawspanlet = R_DrawSpanletConstant66;
			else
				r_polydesc.drawspanlet = R_DrawSpanletConstant33;
		}
		else
		{

			/*
			** choose the correct spanlet routine based on alpha
			*/
			if ( alpha == 1 )
			{

				if ( r_polynonlit )
				{
					r_polydesc.drawspanlet = R_DrawSpanletNonlit;
				}
				else
				{
					// isturbulent is ignored because we know that turbulent surfaces
					// can't be opaque
					r_polydesc.drawspanlet = R_DrawSpanletOpaque;
				}
			}
			else
			{
				if ( r_polynonlit )
				{
					s_spanletvars.alphalevel = alpha * 100;
					r_polydesc.drawspanlet = R_DrawSpanletNonlit_Alpha;
				}
				else if ( sw_stipplealpha->value )
				{
					if ( isturbulent )
					{
						if ( alpha > 0.33 )
							r_polydesc.drawspanlet = R_DrawSpanletTurbulentStipple66;
						else
							r_polydesc.drawspanlet = R_DrawSpanletTurbulentStipple33;
					}
					else
					{
						if ( alpha > 0.33 )
							r_polydesc.drawspanlet = R_DrawSpanlet66Stipple;
						else
							r_polydesc.drawspanlet = R_DrawSpanlet33Stipple;
					}
				}
				else
				{
					if ( isturbulent )
					{
						if ( alpha > 0.33 )
							r_polydesc.drawspanlet = R_DrawSpanletTurbulentBlended66;
						else
							r_polydesc.drawspanlet = R_DrawSpanletTurbulentBlended33;
					}
					else
					{
						if ( alpha > 0.33 )
							r_polydesc.drawspanlet = R_DrawSpanlet66;
						else
							r_polydesc.drawspanlet = R_DrawSpanlet33;
					}
				}
			}
		}
	}

	// clip to the frustum in worldspace
	nump = r_polydesc.nump;
	clip_current = 0;

	for (i=0; i<4; i++)
	{
		nump = R_ClipPolyFace ( nump, &view_clipplanes[i] );
		if (nump < 3)
			return;
		if (nump > MAXWORKINGVERTS)
			ri.Sys_Error( ERR_DROP, "R_ClipAndDrawPoly: too many points: %d", nump );
	}

// transform vertices into viewspace and project
	pv = &r_clip_verts[clip_current][0][0];

	for (i=0; i<nump; i++)
	{
		VectorSubtract ( pv, r_origin, local );
		TransformVector ( local, transformed );

		if (transformed[2] < NEAR_CLIP)
			transformed[2] = NEAR_CLIP;

		pout = &outverts[i];
		pout->zi = 1.0 / transformed[2];

		pout->s = pv[3];
		pout->t = pv[4];

		scale = xscale * pout->zi;
		pout->u = (xcenter + scale * transformed[0]);

		scale = yscale * pout->zi;
		pout->v = (ycenter - scale * transformed[1]);

		pv += sizeof (vec5_t) / sizeof (*pv);
	}

// draw it
	r_polydesc.nump = nump;
	r_polydesc.pverts = outverts;

	R_DrawPoly( isturbulent );
}

static unsigned short *r_16bitscratch = NULL;
static image_t * r_last16bitscratch = NULL;
static void * R_RasterizeTexture16bit( image_t * image )
{
	unsigned short * pdest;
	int x, y;
	byte * psrc;
	byte * pal;
	byte * palptr;

	if ( !r_16bitscratch )
	{
		r_16bitscratch = (unsigned short*)malloc( 1 << ( 8 + 8 + 1 ) );
	}
	if ( r_last16bitscratch == image )
		return r_16bitscratch;
	r_last16bitscratch = image;

	psrc = image->pixels[0];
	pal = image->palette;
	pdest = r_16bitscratch;
	for ( y = 0; y < image->height; y++ )
		for ( x = 0; x < image->width; x++ )
		{
			int r, g, b;

			palptr = pal + ( (*psrc++) << 2 );
			r = sw_state.gammatable[*palptr++];
			g = sw_state.gammatable[*palptr++];
			b = sw_state.gammatable[*palptr++];

			*pdest++ = COLOR16BIT( r, g, b );
		}
	return r_16bitscratch;
}

/*
** R_BuildPolygonFromSurface
*/
static void R_BuildPolygonFromSurface(msurface_t *fa)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	float		*vec;
	vec5_t     *pverts;
	float       tmins[2] = { 0, 0 };

	r_polydesc.nump = 0;

	// reconstruct the polygon
	pedges = currentmodel->edges;
	lnumverts = fa->numedges;

	pverts = r_clip_verts[0];

	for (i=0 ; i<lnumverts ; i++)
	{
		lindex = currentmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = currentmodel->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = currentmodel->vertexes[r_pedge->v[1]].position;
		}

		if ( currentmodel != r_worldmodel )
		{
			vec3_t vert;

			VectorCopy (vec, vert );
			R_EntityRotate (vert);
			VectorAdd (vert, r_entorigin, pverts[i] );
		}
		else
		{
			VectorAdd (vec, r_entorigin, pverts[i] );
		}
	}

	VectorCopy( fa->texinfo->vecs[0], r_polydesc.vright );
	VectorCopy( fa->texinfo->vecs[1], r_polydesc.vup );
	VectorCopy( fa->plane->normal, r_polydesc.vpn );
	VectorCopy( modelorg, r_polydesc.viewer_position );

	if ( fa->flags & SURF_PLANEBACK )
	{
		VectorSubtract( vec3_origin, r_polydesc.vpn, r_polydesc.vpn );
	}

    r_polynonlit = false;

    if ( fa->texinfo->flags & (SURF_WARP|SURF_WAVY|SURF_NONLIT) )
	{
		if ( vid.bitdepth > 8 )
		{
			if ( fa->texinfo->flags & SURF_NONLIT )
			{
				float value;
				int nonlit;

				R_GetColormap( fa->texinfo->image );
				r_polycolormap16 = fa->texinfo->image->colormap16;
				value = fa->texinfo->surf->nonlit;

				if ( fa->texinfo->flags & SURF_USECOLOR )
				{
					float r, g, b;

					r = 0.29f * value * fa->texinfo->surf->color[0];
					g = 0.59f * value * fa->texinfo->surf->color[1];
					b = 0.11f * value * fa->texinfo->surf->color[2];

					value = (r+g+b)*2;
				}

				nonlit = ( 1.0f - value ) * 63;

				if ( nonlit < 0 )
					nonlit = 0;

				if ( nonlit > 63 )
					nonlit = 63;

				r_polycolormap16 += (nonlit << 8);

				if ( fa->texinfo->image->pixels_8bit )
					r_polydesc.pixels = fa->texinfo->image->pixels_8bit;
				else
					r_polydesc.pixels = fa->texinfo->image->pixels[0];

				r_polynonlit = true;
				tmins[0] = fa->texturemins[0];
				tmins[1] = fa->texturemins[1];
			}
			else
			{
				if ( fa->texinfo->image->palette )
				{
					r_polydesc.pixels = (byte*)R_RasterizeTexture16bit( fa->texinfo->image );
				}
				else
				{
					r_polydesc.pixels = fa->texinfo->image->pixels[0];
				}
			}
		}
		else
		{
			if ( fa->texinfo->flags & SURF_NONLIT )
			{
				float value;
				int nonlit;

				r_polycolormap = vid.colormap;
				value = fa->texinfo->surf->nonlit;
				if ( fa->texinfo->flags & SURF_USECOLOR )
				{
					float r, g, b;

					r = 0.29f * value * fa->texinfo->surf->color[0];
					g = 0.59f * value * fa->texinfo->surf->color[1];
					b = 0.11f * value * fa->texinfo->surf->color[2];

					value = (r+g+b)*2;
				}

				nonlit = ( 1.0f - value ) * 63;

				if ( nonlit < 0 )
					nonlit = 0;

				if ( nonlit > 63 )
					nonlit = 63;

				r_polycolormap += (nonlit << 8);
				r_polynonlit = true;
				tmins[0] = fa->texturemins[0];
				tmins[1] = fa->texturemins[1];
			}
			r_polydesc.pixels       = fa->texinfo->image->pixels[0];
		}
		r_polydesc.pixel_width  = fa->texinfo->image->width;
		r_polydesc.pixel_height = fa->texinfo->image->height;
	}
	else
	{
		surfcache_t *scache;

		scache = D_CacheSurface( fa, 0 );

		r_polydesc.pixels       = scache->data;
		r_polydesc.pixel_width  = scache->width;
		r_polydesc.pixel_height = scache->height;

		tmins[0] = fa->texturemins[0];
		tmins[1] = fa->texturemins[1];
	}

	lindex = currentmodel->surfedges[fa->firstedge];

	if (lindex > 0)
	{
		r_pedge = &pedges[lindex];
		vec = currentmodel->vertexes[r_pedge->v[0]].position;
	}
	else
	{
		r_pedge = &pedges[-lindex];
		vec = currentmodel->vertexes[r_pedge->v[1]].position;
	}
	r_polydesc.dist = DotProduct( r_polydesc.vpn, vec );

	r_polydesc.s_offset = fa->texinfo->vecs[0][3] - tmins[0];
	r_polydesc.t_offset = fa->texinfo->vecs[1][3] - tmins[1];

	if (fa->texinfo->flags & SURF_TRANSLATE)
	{
		r_polydesc.s_offset += fa->texinfo->surf->delta_s;
		r_polydesc.t_offset += fa->texinfo->surf->delta_t;
	}
	if ( currentmodel != r_worldmodel )
	{
		R_EntityRotate (r_polydesc.vright);
		R_EntityRotate (r_polydesc.vup);
		R_EntityRotate (r_polydesc.vpn);
	}

	r_polydesc.nump = lnumverts;
}

/*
** R_PolygonCalculateGradients
*/
static void R_PolygonCalculateGradients (void)
{
	vec3_t		p_normal, p_saxis, p_taxis;
	float		distinv;

	TransformVector (r_polydesc.vpn, p_normal);
	TransformVector (r_polydesc.vright, p_saxis);
	TransformVector (r_polydesc.vup, p_taxis);

	distinv = 1.0 / (-(DotProduct (r_polydesc.viewer_position, r_polydesc.vpn)) + r_polydesc.dist );

	d_sdivzstepu  =  p_saxis[0] * xscaleinv;
	d_sdivzstepv  = -p_saxis[1] * yscaleinv;
	d_sdivzorigin =  p_saxis[2] - xcenter * d_sdivzstepu - ycenter * d_sdivzstepv;

	d_tdivzstepu  =  p_taxis[0] * xscaleinv;
	d_tdivzstepv  = -p_taxis[1] * yscaleinv;
	d_tdivzorigin =  p_taxis[2] - xcenter * d_tdivzstepu - ycenter * d_tdivzstepv;

	d_zistepu =   p_normal[0] * xscaleinv * distinv;
	d_zistepv =  -p_normal[1] * yscaleinv * distinv;
	d_ziorigin =  p_normal[2] * distinv - xcenter * d_zistepu - ycenter * d_zistepv;

	sadjust = (fixed16_t) ( ( DotProduct( r_polydesc.viewer_position, r_polydesc.vright) + r_polydesc.s_offset ) * 0x10000 );
	tadjust = (fixed16_t) ( ( DotProduct( r_polydesc.viewer_position, r_polydesc.vup   ) + r_polydesc.t_offset ) * 0x10000 );

// -1 (-epsilon) so we never wander off the edge of the texture
	bbextents = (r_polydesc.pixel_width << 16) - 1;
	bbextentt = (r_polydesc.pixel_height << 16) - 1;
}

/*
** R_DrawPoly
**
** Polygon drawing function.  Uses the polygon described in r_polydesc
** to calculate edges and gradients, then renders the resultant spans.
**
** This should NOT be called externally since it doesn't do clipping!
*/
static void R_DrawPoly( qboolean iswater )
{
	int			i, nump;
	float		ymin, ymax;
	emitpoint_t	*pverts;
	espan_t	spans[MAXHEIGHT+1];

	s_polygon_spans = spans;

// find the top and bottom vertices, and make sure there's at least one scan to
// draw
	ymin = 999999.9;
	ymax = -999999.9;
	pverts = r_polydesc.pverts;

	//
	// initialize min and max just in case
	//
	s_minindex = 0;
	s_maxindex = 0;

	for (i=0 ; i<r_polydesc.nump ; i++)
	{
		if (pverts->v < ymin)
		{
			ymin = pverts->v;
			s_minindex = i;
		}

		if (pverts->v > ymax)
		{
			ymax = pverts->v;
			s_maxindex = i;
		}

		pverts++;
	}

	ymin = ceil (ymin);
	ymax = ceil (ymax);

	if (ymin >= ymax)
		return;		// doesn't cross any scans at all

	cachewidth = r_polydesc.pixel_width;
	cacheblock = r_polydesc.pixels;

// copy the first vertex to the last vertex, so we don't have to deal with
// wrapping
	nump = r_polydesc.nump;
	pverts = r_polydesc.pverts;
	pverts[nump] = pverts[0];

	R_PolygonCalculateGradients ();
	R_PolygonScanLeftEdge ();
	R_PolygonScanRightEdge ();

	R_PolygonDrawSpans( s_polygon_spans, iswater );
}

/*
** R_DrawAlphaSurfaces
*/
void R_DrawAlphaSurfaces( void )
{
	msurface_t *s = r_alpha_surfaces;

	while ( s )
	{
		currententity = s->entity;
		currentmodel = currententity->model;

		if (!currentmodel)
			currentmodel = r_worldmodel;

		if ( currentmodel != r_worldmodel )
		{
			vec3_t angles;

			VectorCopy( currententity->angles, angles );
			angles[ YAW ] = -angles[ YAW ];
			angles[ PITCH ] = -angles[ PITCH ];
			AnglesToMat( angles, entity_rotation );
		}

		VectorCopy (currententity->origin, r_entorigin);
		VectorSubtract (r_origin, r_entorigin, modelorg);

		R_BuildPolygonFromSurface( s );

		R_ClipAndDrawPoly( 1 - s->texinfo->surf->translucence, ( s->texinfo->flags & (SURF_WARP|SURF_WAVY) ) != 0, true );

		//
		// draw any consoles for this surface
		//
		R_DrawConsole( currentmodel, s );

		s = s->nextalphasurface;
	}
	
	r_alpha_surfaces = NULL;
}

/*
** R_IMFlatShadedQuad
*/
void R_IMFlatShadedQuad( vec3_t a, vec3_t b, vec3_t c, vec3_t d, float rr, float gg, float bb, float alpha )
{
	vec3_t s0, s1;

	r_polydesc.nump = 4;
	VectorCopy( r_origin, r_polydesc.viewer_position );

	VectorCopy( a, r_clip_verts[0][0] );
	VectorCopy( b, r_clip_verts[0][1] );
	VectorCopy( c, r_clip_verts[0][2] );
	VectorCopy( d, r_clip_verts[0][3] );

	r_clip_verts[0][0][3] = 0;
	r_clip_verts[0][1][3] = 0;
	r_clip_verts[0][2][3] = 0;
	r_clip_verts[0][3][3] = 0;

	r_clip_verts[0][0][4] = 0;
	r_clip_verts[0][1][4] = 0;
	r_clip_verts[0][2][4] = 0;
	r_clip_verts[0][3][4] = 0;

	VectorSubtract( d, c, s0 );
	VectorSubtract( c, b, s1 );
	CrossProduct( s0, s1, r_polydesc.vpn );
	VectorNormalize( r_polydesc.vpn );

	r_polydesc.dist = DotProduct( r_polydesc.vpn, r_clip_verts[0][0] );

	if ( vid.bitdepth > 8 )
	{
		int color;

		color = COLOR16BIT( rr*255, gg*255, bb*255 );
		r_polyblendcolor3[0] = UNPACKRED16BIT( color );
		r_polyblendcolor3[1] = UNPACKGREEN16BIT( color );
		r_polyblendcolor3[2] = UNPACKBLUE16BIT( color );
	}
	else
	{
		int col;

		col = (int)(rr*31) | ( (int)(gg*63) << 5 ) | ( (int)(bb*31) << 11 );
		r_polyblendcolor = d_16to8table[col];
	}

	R_ClipAndDrawPoly( alpha, false, false );
}

